# SPDX-License-Identifier: MIT
import struct

from ..utils import *

from .asc import StandardASC
from .asc.base import *

SMC_READ_KEY           = 0x10
SMC_WRITE_KEY          = 0x11
SMC_GET_KEY_BY_INDEX   = 0x12
SMC_GET_KEY_INFO       = 0x13
SMC_INITIALIZE         = 0x17
SMC_NOTIFICATION       = 0x18
SMC_RW_KEY             = 0x20

class SMCMessage(Register64):
    TYPE = 7, 0
    UNK = 11, 8, Constant(0)
    ID = 15, 12

class SMCInitialize(SMCMessage):
    TYPE = 7, 0, Constant(SMC_INITIALIZE)

class SMCGetKeyInfo(SMCMessage):
    TYPE = 7, 0, Constant(SMC_GET_KEY_INFO)
    KEY = 63, 32

class SMCGetKeyByIndex(SMCMessage):
    TYPE = 7, 0, Constant(SMC_GET_KEY_BY_INDEX)
    INDEX = 63, 32

class SMCWriteKey(SMCMessage):
    TYPE = 7, 0, Constant(SMC_WRITE_KEY)
    SIZE = 23, 16
    KEY = 63, 32

class SMCReadKey(SMCMessage):
    TYPE = 7, 0, Constant(SMC_READ_KEY)
    SIZE = 23, 16
    KEY = 63, 32

class SMCReadWriteKey(SMCMessage):
    TYPE = 7, 0, Constant(SMC_RW_KEY)
    RSIZE = 23, 16
    WSIZE = 31, 24
    KEY = 63, 32

class SMCResult(Register64):
    RESULT = 7, 0
    ID = 15, 12
    SIZE = 31, 16
    VALUE = 63, 32

class SMCError(Exception):
    pass

class SMCEndpoint(ASCBaseEndpoint):
    BASE_MESSAGE = SMCMessage
    SHORT = "smcep"
    TYPE_MAP = {
        "ui64": ("<Q", None),
        "ui32": ("<I", None),
        "ui16": ("<H", None),
        "ui8 ": ("<B", None),
        "si64": ("<q", None),
        "si32": ("<i", None),
        "si16": ("<h", None),
        "si8 ": ("<b", None),
        "flag": ("<B", None),
        "flt ": ("<f", None),
        "hex_": (None, hexdump),
        "ch8*": (None, lambda c: c.split(b"\x00")[0]),
        "ioft": ("<Q", None),
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.shmem = None
        self.msgid = 0
        self.outstanding = set()
        self.ret = {}

    def start(self):
        self.send(SMCInitialize(ID = 0))
        self.msgid += 1 # important!
        while self.shmem is None:
            self.asc.work()

    def new_msgid(self):
        mid = (self.msgid & 0xF)
        self.msgid += 1
        assert(mid not in self.outstanding)
        self.outstanding.add(mid)
        return mid

    def cmd(self, cmd):
        cmd.ID = self.new_msgid()
        self.send(cmd)
        while cmd.ID in self.outstanding:
            self.asc.work()
        ret = self.ret[cmd.ID]
        if ret.RESULT != 0:
            raise SMCError(f"SMC error {ret}", ret)
        return ret

    def write(self, key, data):
        key = int.from_bytes(key.encode("ascii"), byteorder="big")
        self.asc.iface.writemem(self.shmem, data)
        self.cmd(SMCWriteKey(KEY = key, SIZE = len(data)))

    def read(self, key, size):
        key = int.from_bytes(key.encode("ascii"), byteorder="big")
        ret = self.cmd(SMCReadKey(KEY = key, SIZE = size))
        if size <= 4:
            return struct.pack("<I", ret.VALUE)[:size]
        else:
            return self.asc.iface.readmem(self.shmem, ret.SIZE)

    def rw(self, key, data, outsize):
        key = int.from_bytes(key.encode("ascii"), byteorder="big")
        self.asc.iface.writemem(self.shmem, data)
        ret = self.cmd(SMCReadWriteKey(KEY=key, RSIZE=outsize, WSIZE=len(data)))
        if outsize <= 4:
            return struct.pack("<I", ret.VALUE)[:outsize]
        else:
            return self.asc.iface.readmem(self.shmem, ret.SIZE)

    def get_key_by_index(self, index):
        ret = self.cmd(SMCGetKeyByIndex(INDEX = index))
        key = ret.VALUE.to_bytes(4, byteorder="little").decode("ascii")
        return key

    def get_key_info(self, key):
        key = int.from_bytes(key.encode("ascii"), byteorder="big")
        ret = self.cmd(SMCGetKeyInfo(KEY = key))
        info = self.asc.iface.readmem(self.shmem, 6)
        length, type, flags = struct.unpack("B4sB", info)
        return length, type.decode("ascii"), flags

    def read64(self, key):
        return struct.unpack("<Q", self.read(key, 8))[0]

    def read32(self, key):
        return struct.unpack("<I", self.read(key, 4))[0]

    def read16(self, key):
        return struct.unpack("<H", self.read(key, 2))[0]

    def read8(self, key):
        return struct.unpack("<B", self.read(key, 1))[0]

    def read32b(self, key):
        return struct.unpack(">I", self.read(key, 4))[0]

    def read32f(self, key):
        return struct.unpack("<f", self.read(key, 4))[0]

    def write64(self, key, data):
        self.write(key, struct.pack("<Q", data))

    def write32(self, key, data):
        self.write(key, struct.pack("<I", data))

    def write16(self, key, data):
        self.write(key, struct.pack("<H", data))

    def write8(self, key, data):
        self.write(key, struct.pack("<B", data))

    def rw32(self, key, data):
        return struct.unpack("<I", self.rw(key, struct.pack("<I", data), 4))[0]

    def read_type(self, key, size, typecode):
        fmt, func = self.TYPE_MAP.get(typecode, (None, None))

        val = self.read(key, size)

        if fmt:
            val = struct.unpack(fmt, val)[0]
        if func:
            val = func(val)
        return val

    def handle_msg(self, msg0, msg1):
        if self.shmem is None:
            self.log("Starting up")
            self.shmem = msg0
        else:
            msg = SMCResult(msg0)
            ret = msg.RESULT
            mid = msg.ID
            if ret == SMC_NOTIFICATION:
                self.log(f"Notification: {msg.VALUE:#x}")
                return True
            #print(f"msg {mid} return value {ret}")
            self.outstanding.discard(mid)
            self.ret[mid] = msg

        return True

class SMCClient(StandardASC):
    ENDPOINTS = {
        0x20: SMCEndpoint,
    }



# SPDX-License-Identifier: MIT

import struct
from io import BytesIO
from construct import *
from ..common import *
from ...utils import *
from ..asc import StandardASC
from ..asc.base import *
from .rbep import AFKRingBufEndpoint

EPICType = "EPICType" / Enum(Int32ul,
    NOTIFY = 0,
    COMMAND = 3,
    REPLY = 4,
    NOTIFY_ACK = 8,
)

EPICCategory = "EPICCategory" / Enum(Int8ul,
    REPORT = 0x00,
    NOTIFY = 0x10,
    REPLY = 0x20,
    COMMAND = 0x30,
)

EPICSubtype = "EPICSubtype" / Enum(Int16ul,
    ANNOUNCE = 0x30,
    TEARDOWN = 0x32,
    RETCODE_WITH_PAYLOAD = 0xa0,
    STD_SERVICE = 0xc0,
    RETCODE = 0x84,
    STRING = 0x8a,
)

EPICHeader = Struct(
    "channel" / Int32ul,
    "type" / EPICType,
    "version" / Const(2, Int8ul),
    "seq" / Int16ul,
    "pad" / Const(0, Int8ul),
    "unk" / Const(0, Int32ul),
    "timestamp" / Default(Int64ul, 0),
)

EPICSubHeader = Struct(
    "length" / Int32ul,
    "version" / Default(Int8ul, 4),
    "category" / EPICCategory,
    "type" / EPICSubtype,
    "timestamp" / Default(Int64ul, 0),
    "seq" / Int16ul,
    "unk" / Default(Hex(Int16ul), 0),
    "inline_len" / Hex(Int32ul),
)

EPICSubHeaderV2 = Struct(
    "length" / Int32ul,
    "version" / Default(Int8ul, 2),
    "category" / EPICCategory,
    "type" / EPICSubtype,
    "seq" / Int16ul,
    "unk" / Default(Hex(Int16ul), 0),
    "pad" / Default(Hex(Int64ul), 0),
    "inline_len" / Hex(Int32ul),
)

# dcp's announce
EPICAnnounce = Struct(
    "name" / Padded(32, CString("utf8")),
    "props" / Optional(OSSerialize())
)

# aop's announce
EPICServiceAnnounce = Struct(
    "name" / Padded(20, CString("utf8")),
    "unk1" / Hex(Int32ul),
    "retcode" / Hex(Int32ul), # 0xE00002C2
    "unk3" / Hex(Int32ul),
    "channel" / Hex(Int32ul),
    "unk5" / Hex(Int32ul),
    "unk6" / Hex(Int32ul),
)

EPICSetProp = Struct(
    "name_len" / Int32ul,
    "name" / Aligned(4, CString("utf8")),
    "value" / OSSerialize()
)

EPICCmd = Struct(
    "retcode" / Default(Hex(Int32ul), 0),
    "rxbuf" / Hex(Int64ul),
    "txbuf" / Hex(Int64ul),
    "rxlen" / Hex(Int32ul),
    "txlen" / Hex(Int32ul),
    "rxcookie" / Optional(Default(Bool(Int8ul), False)),
    "txcookie" / Optional(Default(Bool(Int8ul), False)),
)

# Register RX report handlers with @report_handler(type, ConstructClass)
def report_handler(subtype, repcls):
    def f(x):
        x.is_report = True
        x.subtype = subtype
        x.repcls = repcls
        return x
    return f

class EPICError(Exception):
    pass

class EPICService:
    RX_BUFSIZE = 0x4000
    TX_BUFSIZE = 0x4000

    def __init__(self, ep):
        self.iface = ep.asc.iface
        self.ep = ep
        self.ready = False
        self.chan = None
        self.seq = 0

        self.reporthandler = {}
        self.reporttypes = {}
        for name in dir(self):
            i = getattr(self, name)
            if not callable(i):
                continue
            if not getattr(i, "is_report", False):
                continue
            self.reporthandler[i.subtype] = i
            self.reporttypes[i.subtype] = i.repcls

    def log(self, msg):
        print(f"[{self.ep.name}.{self.SHORT}] {msg}")

    def init(self, props):
        self.log(f"Init: {props}")
        self.props = props
        self.rxbuf, self.rxbuf_dva = self.ep.asc.ioalloc(self.RX_BUFSIZE)
        self.txbuf, self.txbuf_dva = self.ep.asc.ioalloc(self.TX_BUFSIZE)
        self.ready = True

    def wait(self):
        while not self.ready:
            self.ep.asc.work()

    def handle_report(self, category, type, seq, fd):
        st = int(type)
        self.log(f"Report {category}/{st:#x} #{seq}")
        handler = self.reporthandler.get(st, None)
        if handler is None:
            self.log("unknown report: 0x%x" % (st))
            chexdump(fd.read())
            return False

        payload = fd.read()
        reptype = self.reporttypes.get(st, None)
        try:
            rep = reptype.parse(payload)
        except Exception as e:
            self.log(e)
            rep = None
            chexdump(payload)
            raise EPICError(f"failed to parse report {st:#x}")
        return handler(seq, fd, rep)

    def handle_notify(self, category, type, seq, fd):
        retcode = struct.unpack("<I", fd.read(4))[0]
        self.log(f"Notify {category}/{type} #{seq} ({retcode})")
        data = fd.read()
        chexdump(data)
        print("Send ACK")

        data = data[:0x50] + b"\x01\x00\x00\x00" + data[0x54:]

        pkt = struct.pack("<I", 0) + data
        self.ep.send_epic(self.chan, EPICType.NOTIFY_ACK, EPICCategory.REPLY, type, seq, pkt, len(data))

    def handle_reply(self, category, type, seq, fd):
        off = fd.tell()
        data = fd.read()

        ret = data
        if (hasattr(self, "last_call") and hasattr(self.last_call, "RETS")):
            call = getattr(self, "last_call")
            item = call.RETS.parse(data)
            ret = item
            self.reply = ret
            return

        if (type == EPICSubtype.RETCODE):
            rc = struct.unpack("<I", data)[0]
            self.log(f"Retcode #{seq}: {rc:#x}")
            self.reply = ret
            return
        elif (type == EPICSubtype.STD_SERVICE):
            fd.seek(off)
            cmd = EPICCmd.parse_stream(fd)
            payload = fd.read()
            self.log(f"Response {int(type):#x} #{seq}: {cmd.retcode:#x}")
            if cmd.retcode != 0:
                raise EPICError(f"IOP returned errcode {cmd.retcode:#x}")
            if payload:
                self.log("Inline payload:")
                chexdump(payload)
            assert cmd.rxbuf == self.rxbuf_dva
            self.reply = self.iface.readmem(self.rxbuf, cmd.rxlen)
            return
        elif (type == EPICSubtype.RETCODE_WITH_PAYLOAD):
            rc = struct.unpack("<I", data[:4])[0]
            self.log(f"Retcode #{seq}: {rc:#x}")
            assert(rc == 0)
            self.reply = ret
            return
        else:
            self.log(f"#{seq}: Unknown reply type: {int(type):#x}")
            chexdump(data)
            self.reply = ret
            return

    def handle_cmd(self, category, type, seq, fd):
        cmd = EPICCmd.parse_stream(fd)
        self.log(f"Command {type:#x} #{seq}: {cmd.retcode:#x}")

    def send_cmd(self, type, data, retlen=None):
        if retlen is None:
            retlen = len(data)
        cmd = Container()
        cmd.rxbuf = self.rxbuf_dva
        cmd.txbuf = self.txbuf_dva
        cmd.txlen = len(data)
        cmd.rxlen = retlen
        self.iface.writemem(self.txbuf, data)
        self.reply = None
        pkt = EPICCmd.build(cmd)
        self.ep.send_epic(self.chan, EPICType.COMMAND, EPICCategory.COMMAND, type, self.seq, pkt)
        self.seq += 1
        while self.reply is None:
            self.ep.asc.work()
        return self.reply

    def send_notify(self, type, data, **kwargs):
        self.reply = None
        self.iface.writemem(self.txbuf, data)
        self.ep.send_epic(self.chan, EPICType.NOTIFY, EPICCategory.NOTIFY, type, self.seq, data, **kwargs)
        self.seq += 1
        while self.reply is None:
            self.ep.asc.work()
        return self.reply

class EPICStandardService(EPICService):
    def call(self, group, cmd, data=b'', replen=None):
        msg = struct.pack("<2xHIII48x", group, cmd, len(data), 0x69706378) + data
        if replen is not None:
            replen += 64
        resp = self.send_cmd(0xc0, msg, replen)
        if not resp:
            return
        rgroup, rcmd, rlen, rmagic = struct.unpack("<2xHIII", resp[:16])
        assert rmagic == 0x69706378
        assert rgroup == group
        assert rcmd == cmd
        return resp[64:64+rlen]

    def getLocation(self, unk=0):
        return struct.unpack("<16xI12x", self.call(4, 4, bytes(32)))

    def getUnit(self, unk=0):
        return struct.unpack("<16xI12x", self.call(4, 5, bytes(32)))

    def open(self, unk=0):
        self.call(4, 6, struct.pack("<16xI12x", unk))

    def close(self):
        self.call(4, 7, bytes(16))

class AFKSystemService(EPICService):
    NAME = "system"
    SHORT = "system"

    def getProperty(self, prop, val):
        pass
        #self.send_cmd(0x40, msg, 0)

    def setProperty(self, prop, val):
        msg = {
            "name_len": (len(prop) + 3) & ~3,
            "name": prop,
            "value": val,
        }
        msg = EPICSetProp.build(msg)
        self.send_cmd(0x43, msg, 0)

class EPICEndpoint(AFKRingBufEndpoint):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.serv_map = {}
        self.chan_map = {}
        self.serv_names = {}
        self.hseq = 0

        for i in self.SERVICES:
            self.serv_names[i.NAME] = i

    def handle_ipc(self, data):
        fd = BytesIO(data)
        hdr = EPICHeader.parse_stream(fd)
        sub = EPICSubHeader.parse_stream(fd)

        if self.verbose > 2:
            self.log(f"Ch {hdr.channel} Type {hdr.type} Ver {hdr.version} Seq {hdr.seq}")
            self.log(f"  Len {sub.length} Ver {sub.version} Cat {sub.category} Type {sub.type:#x} Seq {sub.seq}")

        if sub.category == EPICCategory.REPORT:
            self.handle_report(hdr, sub, fd)
        if sub.category == EPICCategory.NOTIFY:
            self.handle_notify(hdr, sub, fd)
        elif sub.category == EPICCategory.REPLY:
            self.handle_reply(hdr, sub, fd)
        elif sub.category == EPICCategory.COMMAND:
            self.handle_cmd(hdr, sub, fd)

    def wait_for(self, name):
        while True:
            srv = getattr(self, name, None)
            if srv is not None and srv.ready:
                break
            self.asc.work()

    def handle_report(self, hdr, sub, fd):
        if sub.type == EPICSubtype.ANNOUNCE: # dcp's announce
            init = EPICAnnounce.parse_stream(fd)
            if init.props is None:
                init.props = {}
            name = init.name
            if "EPICName" in init.props:
                name = init.props["EPICName"]
            key = name + str(init.props.get("EPICUnit", ""))
            if name in self.serv_names:
                srv = self.serv_names[name](self)
                short = srv.SHORT + str(init.props.get("EPICUnit", ""))
                setattr(self, short, srv)
                srv.init(init.props)
                srv.chan = hdr.channel
                self.chan_map[hdr.channel] = srv
                self.serv_map[key] = srv
                self.log(f"New service: {key} on channel {hdr.channel} (short name: {short})")
            else:
                self.log(f"Unknown service {key} on channel {hdr.channel}")
        elif sub.type == EPICSubtype.STD_SERVICE: # aop's announce
            props = EPICServiceAnnounce.parse_stream(fd)
            name = props.name
            key = props.name
            chan = props.channel # aop uses channel parsed from prop, not from header
            if name in self.serv_names:
                srv = self.serv_names[name](self)
                short = srv.SHORT # aop has no EPICUnit stuff
                #setattr(self, short, srv)
                srv.init(props)
                srv.chan = chan
                self.chan_map[chan] = srv
                self.serv_map[key] = srv
                self.log(f"New service: {key} on channel {chan} (short name: {short})")
            else:
                self.log(f"Unknown service {key} on channel {chan}")
        else:
            if hdr.channel not in self.chan_map:
                self.log(f"Ignoring report on channel {hdr.channel}")
            else:
                self.chan_map[hdr.channel].handle_report(sub.category, sub.type, sub.seq, fd)

    def handle_notify(self, hdr, sub, fd):
        self.chan_map[hdr.channel].handle_notify(sub.category, sub.type, sub.seq, fd)

    def handle_reply(self, hdr, sub, fd):
        self.chan_map[hdr.channel].handle_reply(sub.category, sub.type, sub.seq, fd)

    def handle_cmd(self, hdr, sub, fd):
        self.chan_map[hdr.channel].handle_cmd(sub.category, sub.type, sub.seq, fd)

    def send_roundtrip(self, chan, call, **kwargs):
        self.serv_map[chan].last_call = call
        ret = self.serv_map[chan].send_notify(call.TYPE, call.ARGS.build(call.args), **kwargs)
        self.serv_map[chan].last_call = None
        return ret

    def send_notify(self, chan, call, **kwargs):
        return self.serv_map[chan].send_notify(call.TYPE, call.ARGS.build(call.args), **kwargs)

    def send_cmd(self, chan, type, data, retlen=None, **kwargs):
        return self.serv_map[chan].send_cmd(type, data, retlen, **kwargs)

    def send_epicv4(self, chan, ptype, category, type, seq, data, inline_len=0, **kwargs):
        hdr = Container()
        hdr.channel = chan
        hdr.type = ptype
        hdr.seq = self.hseq
        self.hseq += 1

        sub = Container()
        sub.length = len(data)
        sub.category = category
        sub.type = type
        sub.seq = seq
        sub.inline_len = inline_len
        pkt = EPICHeader.build(hdr) + EPICSubHeader.build(sub) + data
        super().send_ipc(pkt)

    def send_epicv2(self, chan, ptype, category, type, seq, data, inline_len=0, **kwargs):
        hdr = Container()
        hdr.channel = chan
        hdr.type = ptype
        hdr.seq = self.hseq
        self.hseq += 1

        sub = Container()
        sub.length = len(data)
        sub.category = category
        sub.type = type
        sub.seq = seq
        sub.inline_len = inline_len
        pkt = EPICHeader.build(hdr) + EPICSubHeaderV2.build(sub) + data
        super().send_ipc(pkt)

    def send_epic(self, chan, ptype, category, type, seq, data, inline_len=0, version=4, **kwargs):
        if (version == 2):
            return self.send_epicv2(chan, ptype, category, type, seq, data, inline_len, **kwargs)
        if (version == 4):
            return self.send_epicv4(chan, ptype, category, type, seq, data, inline_len, **kwargs)
        raise EPICError(f"unknown epic sub version: {version}")

class AFKSystemEndpoint(EPICEndpoint):
    SHORT = "system"

    SERVICES = [
        AFKSystemService,
    ]



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, pathlib
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

from m1n1.setup import *
from m1n1.fw.smc import SMCClient

smc_addr = u.adt["arm-io/smc"].get_reg(0)[0]
smc = SMCClient(u, smc_addr, None)

smc.start()
smc.start_ep(0x20)

smc.smcep.write32("gP0d", 0x800001)
smc.smcep.write32("gP1a", 1)

smc.stop()



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT

from construct import *
from m1n1.constructutils import ConstructClass

__all__ = []

DRM_COMMAND_BASE = 0x40

ASAHI_BO_PIPELINE = 1

class drm_asahi_submit_t(ConstructClass):
    subcon = Struct(
        "cmdbuf" / Int64ul,
        "in_syncs" / Int64ul,
        "in_sync_count" / Int32ul,
        "out_sync" / Int32ul,
    )

class drm_asahi_wait_bo_t(ConstructClass):
    subcon = Struct(
        "handle" / Int32ul,
        Padding(4),
        "timeout_ns" / Int64sl,
    )

class drm_asahi_create_bo_t(ConstructClass):
    subcon = Struct(
        "size" /  Int32ul,
        "flags" / Int32ul,
        "handle" / Int32ul,
        Padding(4),
        "offset" / Int64ul,
    )

#class drm_asahi_mmap_bo_t(ConstructClass):
    #subcon = Struct(
        #"handle" / Int32ul,
        #"flags" / Int32ul,
        #"offset" / Int64ul,
    #)

class drm_asahi_get_param_t(ConstructClass):
    subcon = Struct(
        "param" / Int32ul,
        Padding(4),
        "value" / Int64ul,
    )

class drm_asahi_get_bo_offset_t(ConstructClass):
    subcon = Struct(
        "handle" / Int32ul,
        Padding(4),
        "offset" / Int64ul,
    )

ASAHI_MAX_ATTACHMENTS = 16

ASAHI_ATTACHMENT_C  = 0
ASAHI_ATTACHMENT_Z  = 1
ASAHI_ATTACHMENT_S  = 2

class drm_asahi_attachment_t(ConstructClass):
    subcon = Struct(
        "type" / Int32ul,
        "size" / Int32ul,
        "pointer" / Int64ul,
    )

ASAHI_CMDBUF_LOAD_C = (1 << 0)
ASAHI_CMDBUF_LOAD_Z = (1 << 1)
ASAHI_CMDBUF_LOAD_S = (1 << 2)

class drm_asahi_cmdbuf_t(ConstructClass):
    subcon = Struct(
        "flags" / Int64ul,

        "encoder_ptr" / Int64ul,
        "encoder_id" / Int32ul,

        "cmd_ta_id" / Int32ul,
        "cmd_3d_id" / Int32ul,

        "ds_flags" / Int32ul,
        "depth_buffer" / Int64ul,
        "stencil_buffer" / Int64ul,

        "scissor_array" / Int64ul,
        "depth_bias_array" / Int64ul,

        "fb_width" / Int32ul,
        "fb_height" / Int32ul,

        "load_pipeline" / Int32ul,
        "load_pipeline_bind" / Int32ul,

        "store_pipeline" / Int32ul,
        "store_pipeline_bind" / Int32ul,

        "partial_reload_pipeline" / Int32ul,
        "partial_reload_pipeline_bind" / Int32ul,

        "partial_store_pipeline" / Int32ul,
        "partial_store_pipeline_bind" / Int32ul,

        "depth_clear_value" / Float32l,
        "stencil_clear_value" / Int8ul,
        Padding(3),

        "attachments" / Array(ASAHI_MAX_ATTACHMENTS, drm_asahi_attachment_t),
        "attachment_count" / Int32ul,
    )

__all__.extend(k for k, v in globals().items()
               if ((callable(v) or isinstance(v, type)) and v.__module__ == __name__) or isinstance(v, int))



#!/usr/bin/env python3

import sys, pathlib
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

from m1n1.setup import *
from m1n1.hw.dart import DART
from m1n1.utils import *

dart = DART.from_adt(u, "arm-io/dart-dispdfr")

width = 2040
height = 60
stride = 64

fb_size = align_up(width * stride * 4, 8 * 0x4000)
buf = u.memalign(0x4000, fb_size)

colors = [0xDD0000, 0xFE6230, 0xFEF600, 0x00BB00, 0x009BFE, 0x000083, 0x30009B]
for i, color in enumerate(colors):
    lines = width // len(colors)
    offset = i * lines * stride * 4
    for j in range(lines):
        p.memset32(buf + offset + j * stride * 4, color, height * 2)
        p.memset32(buf + offset + j * stride * 4 + height * 2, 0xffffffff^color, height * 2)



iova = dart.iomap(0, buf, fb_size)
buf_mask = u.memalign(0x4000, fb_size)
p.memset32(buf_mask, 0xFFFFFFFF, fb_size)
iova_mask = dart.iomap(0, buf_mask, fb_size)
p.write32(0x228202200, iova_mask)

dart.dump_device(0)

#enable backlight
p.write32(0x228600070,0x8051)
p.write32(0x22860006c,0x229)

#enable fifo and vblank
p.write32(0x228400100, 0x613)

# Color correction magic (idk why but it makes colors actually work)
p.write32(0x228202074, 0x1)
p.write32(0x228202028, 0x10)
p.write32(0x22820202c, 0x1)
p.write32(0x228202020, 0x1)
p.write32(0x228202034, 0x1)


#layer enable
p.write32(0x228204020, 0x1)
p.write32(0x228204068, 0x1)
p.write32(0x2282040b4, 0x1)
p.write32(0x2282040f4, 0x1)
p.write32(0x2282040ac, 0x100000)
p.write32(0x228201038, 0x10001)

#layer size
p.write32(0x228204048, height << 16 | width)
p.write32(0x22820404c, height << 16 | width)
p.write32(0x22820407c, height << 16 | width)
p.write32(0x228204054, height << 16 | width)

#global size
p.write32(0x228201030, height << 16 | width)

#some more color correction
p.write32(0x22820402c, 0x53e4001)


pipe = [
    0x20014038,
    0x2a | (stride * 4),
    0x20014030,
    0x2a | iova,
]

pipe = [0xc0000001 | (len(pipe) << 16)] + pipe

def flush(pipe):
 for i in pipe:
  p.write32(0x2282010c0, i)

flush(pipe)



# SPDX-License-Identifier: MIT
from ..utils import *
from enum import IntEnum

__all__ = ["SGXRegs", "SGXRegsT602X", "SGXInfoRegs", "agx_decode_unit", "R_FAULT_INFO"]

class FAULT_REASON(IntEnum):
    INVALID = 0
    AF_FAULT = 1
    WRITE_ONLY = 2
    READ_ONLY = 3
    NO_ACCESS = 4
    UNK = 5

class R_FAULT_INFO(Register64):
    ADDR        = 63, 30
    SIDEBAND    = 29, 23
    CONTEXT     = 22, 17
    UNIT        = 16, 9
    LEVEL       = 8, 7
    UNK_5       = 6, 5
    READ        = 4
    REASON      = 3, 1, FAULT_REASON
    FAULTED     = 0

class SGXRegs(RegMap):
    FAULT_INFO      = 0x17030, R_FAULT_INFO

class SGXRegsT602X(RegMap):
    FAULT_INFO      = 0xd8c0, R_FAULT_INFO
    FAULT_ADDR      = 0xd8c8, Register64

class SGXInfoRegs(RegMap):
    CORE_MASK_0     = 0x1500, Register32,
    CORE_MASK_1     = 0x1514, Register32,

    ID_00           = 0x4000, Register32,
    ID_04           = 0x4004, Register32,
    ID_08           = 0x4008, Register32,
    ID_0c           = 0x400c, Register32,
    ID_10           = 0x4010, Register32,
    ID_14           = 0x4014, Register32,
    ID_18           = 0x4018, Register32,
    ID_1c           = 0x401c, Register32,

    ID_8024         = 0x8024, Register32,

class UNIT_00(IntEnum):
    DCMPn           = 0x00 # VDM/PDM/CDM
    UL1Cn           = 0x01 # VDM/PDM/CDM
    CMPn            = 0x02 # VDM/PDM/CDM
    GSL1_n          = 0x03 # VDM/PDM/CDM
    IAPn            = 0x04 # VDM/PDM/CDM
    VCEn            = 0x05 # VDM
    TEn             = 0x06 # VDM
    RASn            = 0x07 # VDM
    VDMn            = 0x08 # VDM
    PPPn            = 0x09 # VDM
    IPFn            = 0x0a # PDM
    IPF_CPFn        = 0x0b # PDM
    VFn             = 0x0c # PDM
    VF_CPFn         = 0x0d # PDM
    ZLSn            = 0x0e # PDM

class UNIT_A0(IntEnum):
    dPM             = 0xa1 # VDM/PDM/CDM
    dCDM_KS0        = 0xa2 # CDM
    dCDM_KS1        = 0xa3 # CDM
    dCDM_KS2        = 0xa4 # CDM
    dIPP            = 0xa5 # PDM
    dIPP_CS         = 0xa6 # PDM
    dVDM_CSD        = 0xa7 # VDM
    dVDM_SSD        = 0xa8 # VDM
    dVDM_ILF        = 0xa9 # VDM
    dVDM_ILD        = 0xaa # VDM
    dRDE0           = 0xab # VDM/PDM/CDM
    dRDE1           = 0xac # VDM/PDM/CDM
    FC              = 0xad # VDM/PDM/CDM
    GSL2            = 0xae # VDM/PDM/CDM

    GL2CC_META0     = 0xb0 # VDM/PDM/CDM
    GL2CC_META1     = 0xb1 # VDM/PDM/CDM
    GL2CC_META2     = 0xb2 # VDM/PDM/CDM
    GL2CC_META3     = 0xb3 # VDM/PDM/CDM
    GL2CC_META4     = 0xb4 # VDM/PDM/CDM
    GL2CC_META5     = 0xb5 # VDM/PDM/CDM
    GL2CC_META6     = 0xb6 # VDM/PDM/CDM
    GL2CC_META7     = 0xb7 # VDM/PDM/CDM
    GL2CC_MB        = 0xb8 # VDM/PDM/CDM

class UNIT_D0_T602X(IntEnum):
    gCDM_CS         = 0xd0 # CDM
    gCDM_ID         = 0xd1 # CDM
    gCDM_CSR        = 0xd2 # CDM
    gCDM_CSW        = 0xd3 # CDM
    gCDM_CTXR       = 0xd4 # CDM
    gCDM_CTXW       = 0xd5 # CDM
    gIPP            = 0xd6 # PDM
    gIPP_CS         = 0xd7 # PDM
    gKSM_RCE        = 0xd8 # VDM/PDM/CDM

class UNIT_E0_T602X(IntEnum):
    gPM_SPn         = 0xe0 # VDM/PDM/CDM
    gVDM_CSD_SPn    = 0xe1 # VDM
    gVDM_SSD_SPn    = 0xe2 # VDM
    gVDM_ILF_SPn    = 0xe3 # VDM
    gVDM_TFP_SPn    = 0xe4 # VDM
    gVDM_MMB_SPn    = 0xe5 # VDM
    gRDE_SPn        = 0xe6 # VDM/PDM/CDM

class UNIT_E0_T8103(IntEnum):
    gPM_SPn         = 0xe0 # VDM/PDM/CDM
    gVDM_CSD_SPn    = 0xe1 # VDM
    gVDM_SSD_SPn    = 0xe2 # VDM
    gVDM_ILF_SPn    = 0xe3 # VDM
    gVDM_TFP_SPn    = 0xe4 # VDM
    gVDM_MMB_SPn    = 0xe5 # VDM
    gCDM_CS_SPn_KS0 = 0xe6 # CDM
    gCDM_CS_SPn_KS1 = 0xe7 # CDM
    gCDM_CS_SPn_KS2 = 0xe8 # CDM
    gCDM_SPn_KS0    = 0xe9 # CDM
    gCDM_SPn_KS1    = 0xea # CDM
    gCDM_SPn_KS2    = 0xeb # CDM
    gIPP_SPn        = 0xec # PDM
    gIPP_CS_SPn     = 0xed # PDM
    gRDE0_SPn       = 0xee # VDM/PDM/CDM
    gRDE1_SPn       = 0xef # VDM/PDM/CDM

def agx_decode_unit(v):
    if v < 0xa0:
        group = v >> 4
        return UNIT_00(v & 0x0f).name.replace("n", str(group))
    elif v < 0xd0:
        return UNIT_A0(v).name
    elif v < 0xe0:
        return UNIT_D0_T602X(v).name
    else:
        group = (v >> 4) & 1
        return UNIT_E0_T8103(v & 0xef).name.replace("n", str(group))



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, pathlib, time
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

from m1n1.setup import *
from m1n1 import asm

REPETITIONS = 64

PAGE_SIZE = 16384

TEST_ECORE = 1
TEST_PCORE = 4

L2_LINE_SIZE = 128
PNRG_a = 75
PRNG_m = 31337
rnd_idx = 8

def prng(x):
    return (PNRG_a * x) % PRNG_m

SIZE_DATA_ARRAY = (PRNG_m * L2_LINE_SIZE)

data_buf_addr = u.memalign(PAGE_SIZE, SIZE_DATA_ARRAY)
p.memset64(data_buf_addr, 0x5555555555555555, SIZE_DATA_ARRAY)
aop_addr = u.memalign(PAGE_SIZE, PAGE_SIZE)
p.memset64(aop_addr, 0x5555555555555555, PAGE_SIZE)

freq = u.mrs(CNTFRQ_EL0)
code = u.malloc(0x1000)

util = asm.ARMAsm("""
test:
    dc civac, x0
    dc civac, x1
    isb sy

    mov x7, #0x8000
1:
    add x2, x2, #1
    mul x2, x2, x2
    sub x7, x7, #1
    cbnz x7, 1b
    and x2, x2, #(15 << 60)

    add x1, x1, x2
    ldrb w2, [x1, #512]
    and x2, x2, #(15 << 60)

    add x0, x0, x2

    dsb sy
    isb
    mrs x9, S3_2_c15_c0_0 // PMC0_EL1
    isb
    ldr x2, [x0, x2]
    isb
    mrs x10, S3_2_c15_c0_0
    sub x5, x10, x9

    and x2, x2, #(15 << 60)
    mov x7, #0x4000
1:
    add x2, x2, #1
    mul x2, x2, x2
    sub x7, x7, #1
    cbnz x7, 1b

    and x2, x2, #(15 << 60)

    dsb sy
    isb
    mrs x9, S3_2_c15_c0_0
    isb
    ldr x2, [x1, x2]
    isb
    mrs x10, S3_2_c15_c0_0
    sub x0, x10, x9

    isb sy

    lsl x5, x5, #32
    orr x0, x0, x5
    ret
""", code)
for i in util.disassemble():
    print(i)
iface.writemem(code, util.data)
p.dc_cvau(code, len(util.data))
p.ic_ivau(code, len(util.data))

# Set higher cpufreq pstate on all clusters
p.cpufreq_init()
p.smp_start_secondaries()
p.smp_set_wfe_mode(True);

def cpu_call(cpu, x, *args):
    return p.smp_call_sync(cpu, x | REGION_RX_EL1, *args)

def init_core(cpu):
    p.mmu_init_secondary(cpu)

    def mrs(x):
        return u.mrs(x, call=lambda x, *args: cpu_call(cpu, x, *args))
    def msr(x, v):
        u.msr(x, v, call=lambda x, *args: cpu_call(cpu, x, *args))

    is_ecore = not (mrs(MPIDR_EL1) & (1 << 16))
    # Enable DC MVA ops
    v = mrs(EHID4_EL1 if is_ecore else HID4_EL1)
    v &= ~(1 << 11)
    msr(EHID4_EL1 if is_ecore else HID4_EL1, v)

    # Enable PMU
    v = mrs(PMCR0_EL1)
    v |= 1 | (1<<30)
    msr(PMCR0_EL1, v)
    msr(PMCR1_EL1, 0xffffffffffffffff)

    # Enable TBI
    v = mrs(TCR_EL1)
    v |= (1 << 37)
    msr(TCR_EL1, v)

    # Enable user cache ops
    v = mrs(SCTLR_EL1)
    v |= (1 << 26)
    msr(SCTLR_EL1, v)

init_core(TEST_ECORE)
init_core(TEST_PCORE)

# Enable DC MVA ops
v = u.mrs(EHID4_EL1)
v &= ~(1 << 11)
u.msr(EHID4_EL1, v)

def test_cpu(cpu, mask):
    global rnd_idx

    total_aop = total_ptr = 0
    p.memset64(data_buf_addr, 0x5555555555555555, SIZE_DATA_ARRAY)
    p.memset64(aop_addr, 0x5555555555555555, PAGE_SIZE)
    for i in range(REPETITIONS):
        test_offset = L2_LINE_SIZE * rnd_idx
        test_addr = data_buf_addr + test_offset

        p.write64(aop_addr, test_addr | mask | REGION_RWX_EL0)
        p.dc_civac(aop_addr, L2_LINE_SIZE)
        # p.dc_civac(data_buf_addr, SIZE_DATA_ARRAY)

        elapsed = p.smp_call_sync_el0(cpu, util.test | REGION_RWX_EL0, aop_addr | REGION_RWX_EL0, test_addr | REGION_RWX_EL0, 7 << 60)
        time_aop = elapsed >> 32
        time_ptr = elapsed & 0xffffffff
        total_aop += time_aop
        total_ptr += time_ptr

        rnd_idx = prng(rnd_idx)

    return total_aop, total_ptr


print("ECore plain:", test_cpu(TEST_ECORE, 0))
print("ECore mask: ", test_cpu(TEST_ECORE, 0xaaaaaaaa00000000))
print("PCore plain:", test_cpu(TEST_PCORE, 0))
print("PCore mask: ", test_cpu(TEST_PCORE, 0xaaaaaaaa00000000))

for reg in (
    # "HID0_EL1",
    # "HID1_EL1",
    # "HID2_EL1",
    # "HID3_EL1",
    "HID4_EL1",
    # "HID5_EL1",
    # "HID6_EL1",
    # "HID7_EL1",
    # "HID8_EL1",
    # "HID9_EL1",
    # "HID10_EL1",
    "HID11_EL1",
    # "HID13_EL1",
    # "HID14_EL1",
    # "HID16_EL1",
    # "HID17_EL1",
    # "HID18_EL1",
    "HID21_EL1",
    # "HID26_EL1",
    # "HID27_EL1",
):

    cpu = TEST_PCORE
    hid = u.mrs(reg, call=lambda x, *args: cpu_call(cpu, x, *args))

    for i in range(64):
        if (reg, i) not in (
            ("HID4_EL1", 4),
            ("HID11_EL1", 30),
            ("HID21_EL1", 40),
            ):
            continue

        bit = (1 << i)
        print(f"Test {reg} bit {i}:", end=" ")

        u.msr(reg, hid ^ bit, call=lambda x, *args: cpu_call(cpu, x, *args))

        tval = test_cpu(cpu, 0)[1]
        control = test_cpu(cpu, 0xaaaaaaaa00000000)[1]


        if tval < (0.75 * control):
            print(f"DMP active {tval} {control}")
        else:
            print(f"DMP INACTIVE {tval} {control}")

    u.msr(reg, hid, call=lambda x, *args: cpu_call(cpu, x, *args))




#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, pathlib
import time
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

from m1n1.setup import *

p.usb_iodev_vuart_setup(p.iodev_whoami())
p.iodev_set_usage(IODEV.USB_VUART, USAGE.UARTPROXY)



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, pathlib
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

from m1n1.setup import *
from m1n1.hw.dart import DART, DARTRegs
from m1n1.hw.jpeg import *
from m1n1.utils import *
import argparse
import struct
import time
from enum import IntEnum
from PIL import Image, ImageDraw


def divroundup(val, div):
    return (val + div - 1) // div


def yuv2rgb(y, u, v):
    y -= 16
    u -= 128
    v -= 128

    y /= 255
    u /= 255
    v /= 255

    r = y + 1.13983 * v
    g = y - 0.39465 * u - 0.58060 * v
    b = y + 2.03211 * u

    r = min(255, max(0, int(r * 255)))
    g = min(255, max(0, int(g * 255)))
    b = min(255, max(0, int(b * 255)))

    return (r, g, b)


def rgb2yuv(r, g, b):
    r /= 255
    g /= 255
    b /= 255

    y = 0.299*r + 0.587*g + 0.114*b
    u = -0.14713*r - 0.28886*g + 0.436*b
    v = 0.615*r - 0.51499*g - 0.10001*b

    y = y * 255 + 16
    u = u * 255 + 128
    v = v * 255 + 128

    y = min(255, max(0, int(y)))
    u = min(255, max(0, int(u)))
    v = min(255, max(0, int(v)))

    return (y, u, v)


ap = argparse.ArgumentParser(description='JPEG block experiment')
ap.add_argument("--jpeg", dest='which_jpeg', type=str, default='jpeg0',
                help='which JPEG instance (jpeg0/jpeg1)')
g = ap.add_mutually_exclusive_group(required=True)
g.add_argument("-e", "--encode", action='store_true')
g.add_argument("-d", "--decode", action='store_true')
ap.add_argument("--raw-output", type=str, required=False)
ap.add_argument("--decode-scale", type=int, required=False, default=1)
ap.add_argument("--decode-pixelfmt", type=str, required=False, default='RGBA')
ap.add_argument("--decode-rgba-alpha", type=int, required=False, default=255)
ap.add_argument("--encode-subsampling", type=str, required=False, default='444')
ap.add_argument("--encode-rst-interval", type=int, required=False)
ap.add_argument("--encode-pixelfmt", type=str, required=False, default='RGB888')
ap.add_argument("input", type=str)
ap.add_argument("output", type=str)
args = ap.parse_args()

# print(args)

# Perform necessary pre-parsing
if args.decode:
    assert args.decode_scale in [1, 2, 4, 8]
    decode_scale = args.decode_scale
    # FIXME: verify behavior on non-evenly-divisible sizes

    assert args.decode_pixelfmt in [
        'RGBA',
        'BGRA',
        'RGB565',
        'YUV422-CbYCrY',
        'YUV422-YCbYCr',
        'YUV422-planar',
        'YUV420-planar',
        'YUV444-planar',
    ]
    pixfmt = args.decode_pixelfmt

    with open(args.input, 'rb') as f:
        jpeg_data = f.read()

    found_sof0 = False

    jpeg_work = jpeg_data
    while jpeg_work:
        seg_marker = struct.unpack(">H", jpeg_work[:2])[0]
        print(f"Seg {seg_marker:04X}")
        if seg_marker == 0xFFD8:
            # SOI
            jpeg_work = jpeg_work[2:]
        elif seg_marker == 0xFFDA:
            # SOS
            break
        else:
            seg_len = struct.unpack(">H", jpeg_work[2:4])[0]
            assert seg_len >= 2
            seg_data = jpeg_work[4:4 + seg_len - 2]
            jpeg_work = jpeg_work[4 + seg_len - 2:]

            if seg_marker == 0xFFC0:
                # SOF0
                assert not found_sof0
                found_sof0 = True
                sof0 = struct.unpack(">BHHB", seg_data[:6])
                (jpeg_bpp, jpeg_H, jpeg_W, jpeg_components_cnt) = sof0
                # it is not yet verified what the requirements are for inputs
                assert jpeg_bpp == 8
                assert jpeg_components_cnt == 1 or jpeg_components_cnt == 3
                if jpeg_components_cnt == 1:
                    jpeg_MODE = '400'
                else:
                    jpeg_components = {}
                    for i in range(jpeg_components_cnt):
                        comp_id, comp_sampling, _ = seg_data[6+3*i:6+3*(i+1)]
                        jpeg_components[comp_id] = comp_sampling
                    assert 1 in jpeg_components
                    comp_Y = jpeg_components[1]
                    assert 2 in jpeg_components
                    comp_Cb = jpeg_components[2]
                    assert 3 in jpeg_components
                    comp_Cr = jpeg_components[3]

                    if (comp_Y, comp_Cb, comp_Cr) == (0x11, 0x11, 0x11):
                        jpeg_MODE = '444'
                    elif (comp_Y, comp_Cb, comp_Cr) == (0x21, 0x11, 0x11):
                        jpeg_MODE = '422'
                    elif (comp_Y, comp_Cb, comp_Cr) == (0x22, 0x11, 0x11):
                        jpeg_MODE = '420'
                    elif (comp_Y, comp_Cb, comp_Cr) == (0x41, 0x11, 0x11):
                        jpeg_MODE = '411'
                    else:
                        # TODO: 422-vertical, others???
                        # Is it possible to implement them?
                        print("Unsupported subsampling mode")
                        assert False

    assert found_sof0
    print(f"JPEG is {jpeg_W}x{jpeg_H} with subsampling {jpeg_MODE}")

    if jpeg_MODE == '444' or jpeg_MODE == '400':
        macroblock_W, macroblock_H = 8, 8
    elif jpeg_MODE == '422':
        macroblock_W, macroblock_H = 16, 8
    elif jpeg_MODE == '420':
        macroblock_W, macroblock_H = 16, 16
    elif jpeg_MODE == '411':
        macroblock_W, macroblock_H = 32, 8
    else:
        assert False

    # FIXME: Exactly how much extra memory do we need to allocate?
    surface_W = divroundup(jpeg_W // decode_scale, macroblock_W) * macroblock_W
    surface_H = divroundup(jpeg_H // decode_scale, macroblock_H) * macroblock_H
    if pixfmt in ['RGBA', 'BGRA']:
        BYTESPP = 4
    elif pixfmt in ['RGB565', 'YUV422-CbYCrY', 'YUV422-YCbYCr']:
        BYTESPP = 2
    elif pixfmt in ['YUV422-planar', 'YUV420-planar', 'YUV444-planar']:
        BYTESPP = 1
    else:
        assert False
    surface_stride = surface_W * BYTESPP
    surface_sz = surface_stride*surface_H

    if pixfmt == 'YUV422-planar':
        P1_MULW = 1     # FIXME UGLY
        P1_DIVW = 1
        P1_DIVH = 1
    elif pixfmt == 'YUV420-planar':
        P1_MULW = 1
        P1_DIVW = 1
        P1_DIVH = 2
    elif pixfmt == 'YUV444-planar':
        P1_MULW = 2
        P1_DIVW = 1
        P1_DIVH = 1
    if pixfmt in ['YUV422-planar', 'YUV420-planar', 'YUV444-planar']:
        surface_P1_W = surface_W * P1_MULW // P1_DIVW
        surface_P1_H = surface_H // P1_DIVH
        surface_P1_stride = surface_P1_W
        surface_P1_off = surface_sz
        surface_sz += surface_P1_stride*surface_P1_H
    else:
        surface_P1_stride = 0
        surface_P1_off = 0

    input_mem_sz = align_up(len(jpeg_data))
    print(f"Using size {input_mem_sz:08X} for JPEG data")

    output_mem_sz = align_up(surface_sz)
    print(f"Using size {output_mem_sz:08X} for output image")
else:
    assert args.encode_subsampling in ['444', '422', '420', '400']
    if args.encode_subsampling == '444' or args.encode_subsampling == '400':
        macroblock_W, macroblock_H = 8, 8
    elif args.encode_subsampling == '422':
        macroblock_W, macroblock_H = 16, 8
    elif args.encode_subsampling == '420':
        macroblock_W, macroblock_H = 16, 16
    else:
        assert False

    assert args.encode_pixelfmt in [
        'RGB888',
        'RGB101010',
        'RGB565',
        'YUV10',
        'YUV-linear',
        'YUV444-planar',
        'YUV422-planar',
        'YUV420-planar',
    ]
    pixfmt = args.encode_pixelfmt

    # Driver doesn't support this either
    if pixfmt == 'YUV-linear' and args.encode_subsampling == '444':
        print("WARNING: This combination does not appear to work!!!")
    if pixfmt == 'YUV422-planar' and args.encode_subsampling == '444':
        print("WARNING: This combination does not appear to work!!!")
    if pixfmt == 'YUV420-planar' and args.encode_subsampling == '444':
        print("WARNING: This combination does not appear to work!!!")

    image_data = b''
    image_data_P1 = b''
    with Image.open(args.input) as im:
        im_W, im_H = im.size

        if pixfmt != 'YUV420-planar':
            for y in range(im_H):
                for x in range(im_W):
                    r, g, b = im.getpixel((x, y))
                    if pixfmt == 'RGB888':
                        image_data += struct.pack("BBBB", r, g, b, 255)
                    elif pixfmt == 'RGB101010':
                        image_data += struct.pack("<I", (r << 2) | (g << 12) | (b << 22))
                    elif pixfmt == 'RGB565':
                        image_data += struct.pack("<H", (r >> 3) | ((g >> 2) << 5) | ((b >> 3) << 11))
                    elif pixfmt == 'YUV10':
                        # absolute garbage color space conversion
                        # for demonstration purposes only
                        y_, u_, v_ = rgb2yuv(r, g, b)
                        image_data += struct.pack("<I", (y_ << 2) | (u_ << 12) | (v_ << 22))
                    elif pixfmt == 'YUV-linear':
                        # garbage color space conversion, garbage subsampling
                        # for demonstration purposes only
                        y_, u_, v_ = rgb2yuv(r, g, b)
                        if x & 1 == 0:
                            color = u_
                        else:
                            color = v_
                        image_data += struct.pack("BB", y_, color)
                    elif pixfmt == 'YUV444-planar':
                        # garbage color space conversion
                        # for demonstration purposes only
                        y_, u_, v_ = rgb2yuv(r, g, b)
                        image_data += struct.pack("B", y_)
                        image_data_P1 += struct.pack("BB", u_, v_)
                    elif pixfmt == 'YUV422-planar':
                        # garbage color space conversion, garbage subsampling
                        # for demonstration purposes only
                        y_, u_, v_ = rgb2yuv(r, g, b)
                        if x & 1 == 0:
                            color = u_
                        else:
                            color = v_
                        image_data += struct.pack("B", y_)
                        image_data_P1 += struct.pack("B", color)
                    else:
                        assert False
        else:
            for y in range(im_H):
                for x in range(im_W):
                    r, g, b = im.getpixel((x, y))
                    # garbage color space conversion, garbage subsampling
                    # for demonstration purposes only
                    y_, u_, v_ = rgb2yuv(r, g, b)
                    if x & 1 == 0:
                        color = u_
                    else:
                        color = v_
                    image_data += struct.pack("B", y_)
                    if y & 1 == 0:
                        image_data_P1 += struct.pack("B", color)

    if pixfmt in ['RGB888', 'RGB101010', 'YUV10']:
        BYTESPP = 4
        BYTESPP_P1 = 0
        P1_DIVH = 1
    elif pixfmt in ['RGB565', 'YUV-linear']:
        BYTESPP = 2
        BYTESPP_P1 = 0
        P1_DIVH = 1
    elif pixfmt == 'YUV444-planar':
        BYTESPP = 1
        BYTESPP_P1 = 2
        P1_DIVH = 1
    elif pixfmt == 'YUV422-planar':
        BYTESPP = 1
        BYTESPP_P1 = 1
        P1_DIVH = 1
    elif pixfmt == 'YUV420-planar':
        BYTESPP = 1
        BYTESPP_P1 = 1
        P1_DIVH = 2
    else:
        assert False
    surface_stride = im_W * BYTESPP
    surface_sz = surface_stride * im_H
    surface_P1_off = surface_sz
    print(f"Plane 1 offset at {surface_P1_off:08X}")
    surface_P1_stride = im_W * BYTESPP_P1
    surface_sz += surface_P1_stride * im_H // P1_DIVH
    input_mem_sz = align_up(surface_sz)

    output_mem_sz = input_mem_sz

    print(f"Using size {input_mem_sz:08X} for input image")
    print(f"Using size {output_mem_sz:08X} for output data")

# Turn on the JPEG block
p.pmgr_adt_clocks_enable(f'/arm-io/dart-{args.which_jpeg}')
p.pmgr_adt_clocks_enable(f'/arm-io/{args.which_jpeg}')

dart = DART.from_adt(u, f'/arm-io/dart-{args.which_jpeg}')
dart.initialize()

jpeg_base, _ = u.adt[f'/arm-io/{args.which_jpeg}'].get_reg(0)
jpeg = JPEGRegs(u, jpeg_base)


def reset_block():
    jpeg.MODE.val = 0x100
    jpeg.MODE.val = 0x13e

    set_default_regs()

    jpeg.MODE.val = 0x17f
    for _ in range(10000):
        v = jpeg.REG_0x1004.val
        if v == 0:
            break
        print(f"reset 1 -- {v}")
    if (v := jpeg.REG_0x1004.val) != 0:
        print(f"reset 1 failed! -- {v}")
        assert False

    jpeg.RST_INTERVAL.val = 1
    for _ in range(2500):
        v = jpeg.RST_INTERVAL.val
        if v == 1:
            break
        print(f"reset 2 -- {v}")
    if (v := jpeg.RST_INTERVAL.val) != 1:
        print(f"reset 2 failed! -- {v}")
        assert False
    jpeg.RST_INTERVAL.val = 0

    jpeg.ENABLE_RST_LOGGING.val = 0
    jpeg.REG_0x1a8.val = 0
    jpeg.REG_0x1ac.val = 0
    jpeg.REG_0x1b0.val = 0
    jpeg.REG_0x1b4.val = 0
    jpeg.REG_0x1bc.val = 0
    jpeg.REG_0x1c0.val = 0
    jpeg.REG_0x1c4.val = 0
    jpeg.REG_0x1c8.val = 0
    jpeg.REG_0x1cc.val = 0
    jpeg.REG_0x1d0.val = 0
    jpeg.REG_0x1d4.val = 0

    jpeg.MODE.val = 0x143


def set_default_regs(param1=0):
    jpeg.REG_0x0.val = 0
    jpeg.REG_0x0.val = 0
    jpeg.REG_0x4.val = 0
    jpeg.CODEC.val = 0
    jpeg.REG_0x2c.val = 0
    jpeg.REG_0x30.val = 0
    jpeg.REG_0x34.val = 1
    jpeg.REG_0x38.val = 1
    jpeg.CHROMA_HALVE_H_TYPE1.val = 0
    jpeg.CHROMA_HALVE_H_TYPE2.val = 0
    jpeg.CHROMA_HALVE_V_TYPE1.val = 0
    jpeg.CHROMA_HALVE_V_TYPE2.val = 0
    jpeg.CHROMA_DOUBLE_H.val = 0
    jpeg.CHROMA_QUADRUPLE_H.val = 0
    jpeg.CHROMA_DOUBLE_V.val = 0
    jpeg.PLANAR_CHROMA_HALVING.val = 0
    jpeg.PX_USE_PLANE1.val = 0
    jpeg.PX_TILES_W.val = 1
    jpeg.PX_TILES_H.val = 1
    jpeg.PX_PLANE0_WIDTH.val = 1
    jpeg.PX_PLANE0_HEIGHT.val = 1
    jpeg.PX_PLANE0_TILING_H.val = 1
    jpeg.PX_PLANE0_TILING_V.val = 1
    jpeg.PX_PLANE0_STRIDE.val = 1
    jpeg.PX_PLANE1_WIDTH.val = 1
    jpeg.PX_PLANE1_HEIGHT.val = 1
    jpeg.PX_PLANE1_TILING_H.val = 1
    jpeg.PX_PLANE1_TILING_V.val = 1
    jpeg.PX_PLANE1_STRIDE.val = 1
    jpeg.INPUT_START1.val = 0
    jpeg.INPUT_START2.val = 0
    jpeg.REG_0x94.val = 1
    jpeg.REG_0x98.val = 1
    jpeg.INPUT_END.val = 0xffffffff
    jpeg.OUTPUT_START1.val = 0
    jpeg.OUTPUT_START2.val = 0
    jpeg.OUTPUT_END.val = 0xffffffff
    for i in range(11):
        jpeg.MATRIX_MULT[i].val = 0
    for i in range(10):
        jpeg.DITHER[i].val = 0xff
    jpeg.ENCODE_PIXEL_FORMAT.val = 0
    jpeg.ENCODE_COMPONENT0_POS.val = 0
    jpeg.ENCODE_COMPONENT1_POS.val = 0
    jpeg.ENCODE_COMPONENT2_POS.val = 0
    jpeg.ENCODE_COMPONENT3_POS.val = 0
    jpeg.CONVERT_COLOR_SPACE.val = 0
    jpeg.REG_0x118.val = 0
    jpeg.REG_0x11c.val = 0
    jpeg.REG_0x120.val = 0
    jpeg.TILING_ENABLE.val = 0
    jpeg.TILING_PLANE0.val = 0
    jpeg.TILING_PLANE1.val = 0
    jpeg.DECODE_MACROBLOCKS_W.val = 0
    jpeg.DECODE_MACROBLOCKS_H.val = 0
    jpeg.SCALE_FACTOR.val = 0
    jpeg.DECODE_PIXEL_FORMAT.val = 0
    jpeg.YUV422_ORDER.val = 0
    jpeg.RGBA_ORDER.val = 0
    jpeg.RGBA_ALPHA.val = 0
    jpeg.RIGHT_EDGE_PIXELS.val = 0
    jpeg.BOTTOM_EDGE_PIXELS.val = 0
    jpeg.RIGHT_EDGE_SAMPLES.val = 0
    jpeg.BOTTOM_EDGE_SAMPLES.val = 0

    # this is always done on the m1 max hwrev
    jpeg.REG_0x1fc.val = 0
    jpeg.REG_0x200.val = 0
    jpeg.REG_0x204.val = 0
    jpeg.REG_0x208.val = 0
    jpeg.REG_0x214.val = 0
    jpeg.REG_0x218.val = 0
    jpeg.REG_0x21c.val = 0
    jpeg.REG_0x220.val = 0
    jpeg.REG_0x224.val = 0
    jpeg.REG_0x228.val = 0
    jpeg.REG_0x22c.val = 0
    jpeg.REG_0x230.val = 0
    jpeg.REG_0x234.val = 0x1f40
    jpeg.REG_0x244.val = 0
    jpeg.REG_0x248.val = 0
    jpeg.REG_0x258.val = 0
    jpeg.REG_0x25c.val = 0
    jpeg.REG_0x23c.val = 0
    jpeg.REG_0x240.val = 0
    jpeg.REG_0x250.val = 0
    jpeg.REG_0x254.val = 0

    jpeg.REG_0x160.val = param1
    jpeg.TIMEOUT.val = 0
    jpeg.REG_0x20.val = 0xff


print(f"HW revision is {jpeg.HWREV}")
reset_block()

input_buf_phys = u.heap.memalign(0x4000, input_mem_sz)
output_buf_phys = u.heap.memalign(0x4000, output_mem_sz)
print(f"buffers (phys) {input_buf_phys:016X} {output_buf_phys:016X}")

input_buf_iova = dart.iomap(0, input_buf_phys, input_mem_sz)
output_buf_iova = dart.iomap(0, output_buf_phys, output_mem_sz)
print(f"buffers (iova) {input_buf_iova:08X} {output_buf_iova:08X}")
# dart.dump_all()

iface.writemem(input_buf_phys, b'\xAA' * input_mem_sz)
iface.writemem(output_buf_phys, b'\xAA' * output_mem_sz)


if args.decode:
    iface.writemem(input_buf_phys, jpeg_data)
    print("JPEG uploaded")

    jpeg.REG_0x34 = 1
    jpeg.REG_0x2c = 0
    jpeg.REG_0x38 = 0
    if jpeg_MODE == '444':
        jpeg.CODEC.set(CODEC=E_CODEC._444)
    elif jpeg_MODE == '400':
        jpeg.CODEC.set(CODEC=E_CODEC._400)
    elif jpeg_MODE == '422':
        jpeg.CODEC.set(CODEC=E_CODEC._422)
    elif jpeg_MODE == '420':
        jpeg.CODEC.set(CODEC=E_CODEC._420)
    elif jpeg_MODE == '411':
        jpeg.CODEC.set(CODEC=E_CODEC._411)
    else:
        assert False
    if pixfmt == 'RGBA' or pixfmt == 'BGRA':
        jpeg.DECODE_PIXEL_FORMAT.set(FORMAT=E_DECODE_PIXEL_FORMAT.RGBA8888)
    elif pixfmt == 'RGB565':
        jpeg.DECODE_PIXEL_FORMAT.set(FORMAT=E_DECODE_PIXEL_FORMAT.RGB565)
    elif pixfmt == 'YUV422-CbYCrY' or pixfmt == 'YUV422-YCbYCr':
        jpeg.DECODE_PIXEL_FORMAT.set(FORMAT=E_DECODE_PIXEL_FORMAT.YUV422_linear)
    elif pixfmt == 'YUV422-planar':
        jpeg.DECODE_PIXEL_FORMAT.set(FORMAT=E_DECODE_PIXEL_FORMAT.YUV422_planar)
    elif pixfmt == 'YUV420-planar':
        jpeg.DECODE_PIXEL_FORMAT.set(FORMAT=E_DECODE_PIXEL_FORMAT.YUV420_planar)
    elif pixfmt == 'YUV444-planar':
        jpeg.DECODE_PIXEL_FORMAT.set(FORMAT=E_DECODE_PIXEL_FORMAT.YUV444_planar)
    else:
        assert False

    if pixfmt in ['YUV422-planar', 'YUV420-planar', 'YUV444-planar']:
        jpeg.PX_USE_PLANE1 = 1
        jpeg.PX_PLANE1_WIDTH = jpeg_W * P1_MULW // P1_DIVW // decode_scale - 1
        jpeg.PX_PLANE1_HEIGHT = jpeg_H // P1_DIVH // decode_scale - 1
    else:
        jpeg.PX_USE_PLANE1 = 0
    jpeg.PX_PLANE0_WIDTH = jpeg_W*BYTESPP // decode_scale - 1
    jpeg.PX_PLANE0_HEIGHT = jpeg_H // decode_scale - 1
    jpeg.TIMEOUT = 266000000

    jpeg.REG_0x94 = 0x1f
    jpeg.REG_0x98 = 1

    jpeg.DECODE_MACROBLOCKS_W = divroundup(jpeg_W, macroblock_W)
    jpeg.DECODE_MACROBLOCKS_H = divroundup(jpeg_H, macroblock_H)
    right_edge_px = \
        jpeg_W - divroundup(jpeg_W, macroblock_W)*macroblock_W + macroblock_W
    bot_edge_px = \
        jpeg_H - divroundup(jpeg_H, macroblock_H)*macroblock_H + macroblock_H
    # XXX changing this does not seem to do anything.
    # Does it possibly affect scaling down?
    jpeg.RIGHT_EDGE_PIXELS.val = right_edge_px
    jpeg.BOTTOM_EDGE_PIXELS.val = bot_edge_px
    jpeg.RIGHT_EDGE_SAMPLES.val = right_edge_px // (macroblock_W // 8)
    jpeg.BOTTOM_EDGE_SAMPLES.val = bot_edge_px // (macroblock_H // 8)

    jpeg.PX_TILES_H = divroundup(jpeg_H, macroblock_H)
    # FIXME explain this
    if pixfmt in ['RGBA', 'BGRA', 'RGB565', 'YUV444-planar']:
        jpeg.PX_TILES_W = divroundup(jpeg_W // decode_scale, macroblock_W)
    else:
        jpeg.PX_TILES_W = divroundup(jpeg_W // decode_scale, max(macroblock_W, 16))
    if pixfmt == 'RGBA' or pixfmt == 'BGRA':
        if jpeg_MODE == '444' or jpeg_MODE == '400':
            jpeg.PX_PLANE0_TILING_H = 4
            jpeg.PX_PLANE0_TILING_V = 8 // decode_scale
            jpeg.PX_PLANE1_TILING_H = 1
            jpeg.PX_PLANE1_TILING_V = 1
        elif jpeg_MODE == '422':
            jpeg.PX_PLANE0_TILING_H = 8
            jpeg.PX_PLANE0_TILING_V = 8 // decode_scale
            jpeg.PX_PLANE1_TILING_H = 1
            jpeg.PX_PLANE1_TILING_V = 1
        elif jpeg_MODE == '420':
            jpeg.PX_PLANE0_TILING_H = 8
            jpeg.PX_PLANE0_TILING_V = 16 // decode_scale
            jpeg.PX_PLANE1_TILING_H = 0
            jpeg.PX_PLANE1_TILING_V = 0
        elif jpeg_MODE == '411':
            jpeg.PX_PLANE0_TILING_H = 16
            jpeg.PX_PLANE0_TILING_V = 8 // decode_scale
            jpeg.PX_PLANE1_TILING_H = 0
            jpeg.PX_PLANE1_TILING_V = 0
        else:
            assert False
    elif pixfmt == 'RGB565':
        if jpeg_MODE == '444' or jpeg_MODE == '400':
            jpeg.PX_PLANE0_TILING_H = 2
            jpeg.PX_PLANE0_TILING_V = 8 // decode_scale
            jpeg.PX_PLANE1_TILING_H = 1
            jpeg.PX_PLANE1_TILING_V = 1
        elif jpeg_MODE == '422':
            jpeg.PX_PLANE0_TILING_H = 4
            jpeg.PX_PLANE0_TILING_V = 8 // decode_scale
            jpeg.PX_PLANE1_TILING_H = 1
            jpeg.PX_PLANE1_TILING_V = 1
        elif jpeg_MODE == '420':
            jpeg.PX_PLANE0_TILING_H = 4
            jpeg.PX_PLANE0_TILING_V = 16 // decode_scale
            jpeg.PX_PLANE1_TILING_H = 0
            jpeg.PX_PLANE1_TILING_V = 0
        elif jpeg_MODE == '411':
            jpeg.PX_PLANE0_TILING_H = 8
            jpeg.PX_PLANE0_TILING_V = 8 // decode_scale
            jpeg.PX_PLANE1_TILING_H = 0
            jpeg.PX_PLANE1_TILING_V = 0
        else:
            assert False
    elif pixfmt == 'YUV422-CbYCrY' or pixfmt == 'YUV422-YCbYCr':
        if jpeg_MODE == '444' or jpeg_MODE == '400':
            jpeg.PX_PLANE0_TILING_H = 4
            jpeg.PX_PLANE0_TILING_V = 8 // decode_scale
            jpeg.PX_PLANE1_TILING_H = 1
            jpeg.PX_PLANE1_TILING_V = 1
        elif jpeg_MODE == '422':
            jpeg.PX_PLANE0_TILING_H = 4
            jpeg.PX_PLANE0_TILING_V = 8 // decode_scale
            jpeg.PX_PLANE1_TILING_H = 1
            jpeg.PX_PLANE1_TILING_V = 1
        elif jpeg_MODE == '420':
            jpeg.PX_PLANE0_TILING_H = 4
            jpeg.PX_PLANE0_TILING_V = 16 // decode_scale
            jpeg.PX_PLANE1_TILING_H = 0
            jpeg.PX_PLANE1_TILING_V = 0
        elif jpeg_MODE == '411':
            jpeg.PX_PLANE0_TILING_H = 8
            jpeg.PX_PLANE0_TILING_V = 8 // decode_scale
            jpeg.PX_PLANE1_TILING_H = 0
            jpeg.PX_PLANE1_TILING_V = 0
        else:
            assert False
    elif pixfmt == 'YUV422-planar':
        if jpeg_MODE == '444' or jpeg_MODE == '400':
            jpeg.PX_PLANE0_TILING_H = 2
            jpeg.PX_PLANE0_TILING_V = 8 // decode_scale
            jpeg.PX_PLANE1_TILING_H = 2
            jpeg.PX_PLANE1_TILING_V = 8 // decode_scale
        elif jpeg_MODE == '422':
            jpeg.PX_PLANE0_TILING_H = 2
            jpeg.PX_PLANE0_TILING_V = 8 // decode_scale
            jpeg.PX_PLANE1_TILING_H = 2
            jpeg.PX_PLANE1_TILING_V = 8 // decode_scale
        elif jpeg_MODE == '420':
            jpeg.PX_PLANE0_TILING_H = 2
            jpeg.PX_PLANE0_TILING_V = 16 // decode_scale
            jpeg.PX_PLANE1_TILING_H = 2
            jpeg.PX_PLANE1_TILING_V = 16 // decode_scale
        elif jpeg_MODE == '411':
            jpeg.PX_PLANE0_TILING_H = 4
            jpeg.PX_PLANE0_TILING_V = 8 // decode_scale
            jpeg.PX_PLANE1_TILING_H = 4
            jpeg.PX_PLANE1_TILING_V = 8 // decode_scale
        else:
            assert False
    elif pixfmt == 'YUV420-planar':
        if jpeg_MODE == '444' or jpeg_MODE == '400':
            jpeg.PX_PLANE0_TILING_H = 2
            jpeg.PX_PLANE0_TILING_V = 8 // decode_scale
            jpeg.PX_PLANE1_TILING_H = 2
            jpeg.PX_PLANE1_TILING_V = 4 // decode_scale
        elif jpeg_MODE == '422':
            jpeg.PX_PLANE0_TILING_H = 2
            jpeg.PX_PLANE0_TILING_V = 8 // decode_scale
            jpeg.PX_PLANE1_TILING_H = 2
            jpeg.PX_PLANE1_TILING_V = 4 // decode_scale
        elif jpeg_MODE == '420':
            jpeg.PX_PLANE0_TILING_H = 2
            jpeg.PX_PLANE0_TILING_V = 16 // decode_scale
            jpeg.PX_PLANE1_TILING_H = 2
            jpeg.PX_PLANE1_TILING_V = 8 // decode_scale
        elif jpeg_MODE == '411':
            jpeg.PX_PLANE0_TILING_H = 4
            jpeg.PX_PLANE0_TILING_V = 8 // decode_scale
            jpeg.PX_PLANE1_TILING_H = 4
            jpeg.PX_PLANE1_TILING_V = 4 // decode_scale
        else:
            assert False
    elif pixfmt == 'YUV444-planar':
        if jpeg_MODE == '444' or jpeg_MODE == '400':
            jpeg.PX_PLANE0_TILING_H = 1
            jpeg.PX_PLANE0_TILING_V = 8 // decode_scale
            jpeg.PX_PLANE1_TILING_H = 2
            jpeg.PX_PLANE1_TILING_V = 8 // decode_scale
        elif jpeg_MODE == '422':
            # The driver doesn't use this, but guessing seems to be fine?
            jpeg.PX_PLANE0_TILING_H = 2
            jpeg.PX_PLANE0_TILING_V = 8 // decode_scale
            jpeg.PX_PLANE1_TILING_H = 4
            jpeg.PX_PLANE1_TILING_V = 8 // decode_scale
        elif jpeg_MODE == '420':
            # The driver doesn't use this, but guessing seems to be fine?
            jpeg.PX_PLANE0_TILING_H = 2
            jpeg.PX_PLANE0_TILING_V = 16 // decode_scale
            jpeg.PX_PLANE1_TILING_H = 4
            jpeg.PX_PLANE1_TILING_V = 16 // decode_scale
        elif jpeg_MODE == '411':
            # The driver doesn't use this, but guessing seems to be fine?
            jpeg.PX_PLANE0_TILING_H = 4
            jpeg.PX_PLANE0_TILING_V = 8 // decode_scale
            jpeg.PX_PLANE1_TILING_H = 8
            jpeg.PX_PLANE1_TILING_V = 8 // decode_scale
        else:
            assert False
    else:
        assert False

    if pixfmt in ['RGBA', 'BGRA', 'RGB565', 'YUV444-planar']:
        if jpeg_MODE in ['422', '420']:
            jpeg.CHROMA_DOUBLE_H = 1

        if jpeg_MODE == '411':
            jpeg.CHROMA_QUADRUPLE_H = 1

        if jpeg_MODE == '420':
            jpeg.CHROMA_DOUBLE_V = 1
    elif pixfmt in ["YUV422-CbYCrY", "YUV422-YCbYCr", "YUV422-planar"]:
        if jpeg_MODE == '444':
            jpeg.CHROMA_HALVE_H_TYPE1 = 1

        if jpeg_MODE == '411':
            jpeg.CHROMA_DOUBLE_H = 1

        if jpeg_MODE == '420':
            jpeg.CHROMA_DOUBLE_V = 1
    elif pixfmt in ["YUV420-planar"]:
        if jpeg_MODE == '444':
            jpeg.CHROMA_HALVE_H_TYPE1 = 1

        if jpeg_MODE in ['444', '422', '411']:
            jpeg.CHROMA_HALVE_V_TYPE1 = 1

        if jpeg_MODE == '411':
            jpeg.CHROMA_DOUBLE_H = 1
    else:
        assert False

    jpeg.MATRIX_MULT[0].val = 0x100
    jpeg.MATRIX_MULT[1].val = 0x0
    jpeg.MATRIX_MULT[2].val = 0x167
    jpeg.MATRIX_MULT[3].val = 0x100
    jpeg.MATRIX_MULT[4].val = 0xffffffa8
    jpeg.MATRIX_MULT[5].val = 0xffffff49
    jpeg.MATRIX_MULT[6].val = 0x100
    jpeg.MATRIX_MULT[7].val = 0x1c6
    jpeg.MATRIX_MULT[8].val = 0x0
    jpeg.MATRIX_MULT[9].val = 0x0
    jpeg.MATRIX_MULT[10].val = 0xffffff80

    jpeg.RGBA_ALPHA = args.decode_rgba_alpha
    jpeg.RGBA_ORDER = pixfmt == "RGBA"
    jpeg.YUV422_ORDER = pixfmt == "YUV422-YCbYCr"

    if decode_scale == 1:
        jpeg.SCALE_FACTOR.set(SCALE=E_SCALE.DIV1)
    elif decode_scale == 2:
        jpeg.SCALE_FACTOR.set(SCALE=E_SCALE.DIV2)
    elif decode_scale == 4:
        jpeg.SCALE_FACTOR.set(SCALE=E_SCALE.DIV4)
    elif decode_scale == 8:
        jpeg.SCALE_FACTOR.set(SCALE=E_SCALE.DIV8)
    else:
        assert False

    jpeg.INPUT_START1 = input_buf_iova
    jpeg.INPUT_START2 = 0xdeadbeef
    jpeg.INPUT_END = input_buf_iova + input_mem_sz
    jpeg.OUTPUT_START1 = output_buf_iova
    jpeg.OUTPUT_START2 = output_buf_iova + surface_P1_off
    jpeg.OUTPUT_END = output_buf_iova + output_mem_sz
    jpeg.PX_PLANE0_STRIDE = surface_stride
    jpeg.PX_PLANE1_STRIDE = surface_P1_stride

    jpeg.REG_0x1ac = 0x0
    jpeg.REG_0x1b0 = 0x0
    jpeg.REG_0x1b4 = 0x0
    jpeg.REG_0x1bc = 0x0
    jpeg.REG_0x1c0 = 0x0
    jpeg.REG_0x1c4 = 0x0

    jpeg.REG_0x118 = 0x0
    jpeg.REG_0x11c = 0x1

    jpeg.MODE = 0x177
    jpeg.REG_0x1028 = 0x400

    jpeg.JPEG_IO_FLAGS = 0x3f
    jpeg.REG_0x0 = 0x1
    jpeg.REG_0x1004 = 0x1

    # FIXME: we don't actually know when it's done
    time.sleep(1)

    print(jpeg.STATUS.reg)
    print(jpeg.PERFCOUNTER.reg)

    output_data = iface.readmem(output_buf_phys, output_mem_sz)
    if args.raw_output is not None:
        with open(args.raw_output, 'wb') as f:
            f.write(output_data)

    # Just for demonstration purposes, wrangle everything back into RGB
    with Image.new(
            mode='RGBA',
            size=(jpeg_W // decode_scale, jpeg_H // decode_scale)) as im:
        if pixfmt in ["RGBA", "BGRA", "RGB565"]:
            for y in range(jpeg_H // decode_scale):
                for x in range(jpeg_W // decode_scale):
                    block = output_data[
                        y*surface_stride + x*BYTESPP:
                        y*surface_stride + (x+1)*BYTESPP]

                    if pixfmt == "RGBA":
                        r, g, b, a = block
                    elif pixfmt == "BGRA":
                        b, g, r, a = block
                    elif pixfmt == "RGB565":
                        rgb = struct.unpack("<H", block)[0]
                        b = (rgb & 0b11111) << 3
                        g = ((rgb >> 5) & 0b111111) << 2
                        r = ((rgb >> 11) & 0b11111) << 3
                        a = 255
                    else:
                        assert False
                    im.putpixel((x, y), (r, g, b, a))
        elif pixfmt in ["YUV422-CbYCrY", "YUV422-YCbYCr"]:
            for y in range(jpeg_H // decode_scale):
                for x in range(0, jpeg_W // decode_scale, 2):
                    block = output_data[
                        y*surface_stride + x*BYTESPP:
                        y*surface_stride + (x+2)*BYTESPP]

                    if pixfmt == "YUV422-CbYCrY":
                        cb, y0, cr, y1 = block
                    elif pixfmt == "YUV422-YCbYCr":
                        y0, cb, y1, cr = block

                    r0, g0, b0 = yuv2rgb(y0, cb, cr)
                    r1, g1, b1 = yuv2rgb(y1, cb, cr)

                    im.putpixel((x, y), (r0, g0, b0, 255))
                    # XXX this really needs some fixing
                    if x+1 < jpeg_W // decode_scale:
                        im.putpixel((x+1, y), (r1, g1, b1, 255))
        elif pixfmt == "YUV422-planar":
            for y in range(jpeg_H // decode_scale):
                for x in range(jpeg_W // decode_scale):
                    y_ = output_data[y*surface_stride + x]
                    cb = output_data[surface_P1_off + y*surface_P1_stride + x&~1]
                    cr = output_data[surface_P1_off + y*surface_P1_stride + (x&~1)+1]

                    r, g, b = yuv2rgb(y_, cb, cr)

                    im.putpixel((x, y), (r, g, b, 255))
        elif pixfmt == "YUV420-planar":
            for y in range(jpeg_H // decode_scale):
                for x in range(jpeg_W // decode_scale):
                    y_ = output_data[y*surface_stride + x]
                    cb = output_data[surface_P1_off + (y//2)*surface_P1_stride + x&~1]
                    cr = output_data[surface_P1_off + (y//2)*surface_P1_stride + (x&~1)+1]

                    r, g, b = yuv2rgb(y_, cb, cr)

                    im.putpixel((x, y), (r, g, b, 255))
        elif pixfmt == "YUV444-planar":
            for y in range(jpeg_H // decode_scale):
                for x in range(jpeg_W // decode_scale):
                    y_ = output_data[y*surface_stride + x]
                    cb = output_data[surface_P1_off + y*surface_P1_stride + x*2]
                    cr = output_data[surface_P1_off + y*surface_P1_stride + x*2+1]

                    r, g, b = yuv2rgb(y_, cb, cr)

                    im.putpixel((x, y), (r, g, b, 255))
        else:
            assert False
        im.save(args.output)

if args.encode:
    iface.writemem(input_buf_phys, image_data)
    iface.writemem(input_buf_phys + surface_P1_off, image_data_P1)
    print("Pixel data uploaded")

    jpeg.MODE = 0x17f
    jpeg.REG_0x38 = 0x1     # if not set nothing happens
    jpeg.REG_0x2c = 0x1     # if not set only header is output
    jpeg.REG_0x34 = 0x0     # if set output is a JPEG but weird with no footer

    if args.encode_subsampling == '444':
        jpeg.CODEC.set(CODEC=E_CODEC._444)
    elif args.encode_subsampling == '422':
        jpeg.CODEC.set(CODEC=E_CODEC._422)
    elif args.encode_subsampling == '420':
        jpeg.CODEC.set(CODEC=E_CODEC._420)
    elif args.encode_subsampling == '400':
        jpeg.CODEC.set(CODEC=E_CODEC._400)
    else:
        assert False

    if BYTESPP_P1 != 0:
        jpeg.PX_USE_PLANE1 = 1
        jpeg.PX_PLANE1_WIDTH = im_W*BYTESPP_P1 - 1
        jpeg.PX_PLANE1_HEIGHT = im_H // P1_DIVH - 1
    else:
        jpeg.PX_USE_PLANE1 = 0
        jpeg.PX_PLANE1_WIDTH = 0xffffffff
        jpeg.PX_PLANE1_HEIGHT = 0xffffffff
    jpeg.PX_PLANE0_WIDTH = im_W*BYTESPP - 1
    jpeg.PX_PLANE0_HEIGHT = im_H - 1
    jpeg.TIMEOUT = 266000000

    jpeg.PX_TILES_W = divroundup(im_W, macroblock_W)
    jpeg.PX_TILES_H = divroundup(im_H, macroblock_H)
    if pixfmt in ['RGB888', 'RGB101010', 'YUV10']:
        if args.encode_subsampling == '444' or args.encode_subsampling == '400':
            jpeg.PX_PLANE0_TILING_H = 4
            jpeg.PX_PLANE0_TILING_V = 8
            jpeg.PX_PLANE1_TILING_H = 1
            jpeg.PX_PLANE1_TILING_V = 1
        elif args.encode_subsampling == '422':
            jpeg.PX_PLANE0_TILING_H = 8
            jpeg.PX_PLANE0_TILING_V = 8
            jpeg.PX_PLANE1_TILING_H = 1
            jpeg.PX_PLANE1_TILING_V = 1
        elif args.encode_subsampling == '420':
            jpeg.PX_PLANE0_TILING_H = 8
            jpeg.PX_PLANE0_TILING_V = 16
            jpeg.PX_PLANE1_TILING_H = 0
            jpeg.PX_PLANE1_TILING_V = 0
        else:
            assert False
    elif pixfmt == 'RGB565':
        if args.encode_subsampling == '444' or args.encode_subsampling == '400':
            jpeg.PX_PLANE0_TILING_H = 2
            jpeg.PX_PLANE0_TILING_V = 8
            jpeg.PX_PLANE1_TILING_H = 1
            jpeg.PX_PLANE1_TILING_V = 1
        elif args.encode_subsampling == '422':
            jpeg.PX_PLANE0_TILING_H = 4
            jpeg.PX_PLANE0_TILING_V = 8
            jpeg.PX_PLANE1_TILING_H = 1
            jpeg.PX_PLANE1_TILING_V = 1
        elif args.encode_subsampling == '420':
            jpeg.PX_PLANE0_TILING_H = 4
            jpeg.PX_PLANE0_TILING_V = 16
            jpeg.PX_PLANE1_TILING_H = 0
            jpeg.PX_PLANE1_TILING_V = 0
        else:
            assert False
    elif pixfmt == 'YUV-linear':
        if args.encode_subsampling == '444' or args.encode_subsampling == '400':
            jpeg.PX_PLANE0_TILING_H = 2
            jpeg.PX_PLANE0_TILING_V = 8
            jpeg.PX_PLANE1_TILING_H = 1
            jpeg.PX_PLANE1_TILING_V = 1
        elif args.encode_subsampling == '422':
            jpeg.PX_PLANE0_TILING_H = 4
            jpeg.PX_PLANE0_TILING_V = 8
            jpeg.PX_PLANE1_TILING_H = 1
            jpeg.PX_PLANE1_TILING_V = 1
        elif args.encode_subsampling == '420':
            jpeg.PX_PLANE0_TILING_H = 4
            jpeg.PX_PLANE0_TILING_V = 16
            jpeg.PX_PLANE1_TILING_H = 0
            jpeg.PX_PLANE1_TILING_V = 0
        else:
            assert False
    elif pixfmt == 'YUV444-planar':
        if args.encode_subsampling == '444' or args.encode_subsampling == '400':
            jpeg.PX_PLANE0_TILING_H = 1
            jpeg.PX_PLANE0_TILING_V = 8
            jpeg.PX_PLANE1_TILING_H = 2
            jpeg.PX_PLANE1_TILING_V = 8
        elif args.encode_subsampling == '422':
            jpeg.PX_PLANE0_TILING_H = 2
            jpeg.PX_PLANE0_TILING_V = 8
            jpeg.PX_PLANE1_TILING_H = 4
            jpeg.PX_PLANE1_TILING_V = 8
        elif args.encode_subsampling == '420':
            jpeg.PX_PLANE0_TILING_H = 2
            jpeg.PX_PLANE0_TILING_V = 16
            jpeg.PX_PLANE1_TILING_H = 4
            jpeg.PX_PLANE1_TILING_V = 16
        else:
            assert False
    elif pixfmt == 'YUV422-planar':
        if args.encode_subsampling == '444' or args.encode_subsampling == '400':
            jpeg.PX_PLANE0_TILING_H = 1
            jpeg.PX_PLANE0_TILING_V = 8
            jpeg.PX_PLANE1_TILING_H = 1
            jpeg.PX_PLANE1_TILING_V = 8
        elif args.encode_subsampling == '422':
            jpeg.PX_PLANE0_TILING_H = 2
            jpeg.PX_PLANE0_TILING_V = 8
            jpeg.PX_PLANE1_TILING_H = 2
            jpeg.PX_PLANE1_TILING_V = 8
        elif args.encode_subsampling == '420':
            jpeg.PX_PLANE0_TILING_H = 2
            jpeg.PX_PLANE0_TILING_V = 16
            jpeg.PX_PLANE1_TILING_H = 2
            jpeg.PX_PLANE1_TILING_V = 16
        else:
            assert False
    elif pixfmt == 'YUV420-planar':
        if args.encode_subsampling == '444' or args.encode_subsampling == '400':
            jpeg.PX_PLANE0_TILING_H = 1
            jpeg.PX_PLANE0_TILING_V = 8
            jpeg.PX_PLANE1_TILING_H = 1
            jpeg.PX_PLANE1_TILING_V = 4
        elif args.encode_subsampling == '422':
            jpeg.PX_PLANE0_TILING_H = 2
            jpeg.PX_PLANE0_TILING_V = 8
            jpeg.PX_PLANE1_TILING_H = 2
            jpeg.PX_PLANE1_TILING_V = 4
        elif args.encode_subsampling == '420':
            jpeg.PX_PLANE0_TILING_H = 2
            jpeg.PX_PLANE0_TILING_V = 16
            jpeg.PX_PLANE1_TILING_H = 2
            jpeg.PX_PLANE1_TILING_V = 8
        else:
            assert False
    else:
        assert False
    jpeg.PX_PLANE0_STRIDE = surface_stride
    jpeg.PX_PLANE1_STRIDE = surface_P1_stride

    if pixfmt in ['RGB888', 'RGB101010', 'RGB565', 'YUV10', 'YUV444-planar']:
        if args.encode_subsampling in ['422', '420']:
            jpeg.CHROMA_HALVE_H_TYPE1 = 1
        if args.encode_subsampling == '420':
            jpeg.CHROMA_HALVE_V_TYPE1 = 1
    elif pixfmt in ['YUV-linear', 'YUV422-planar']:
        if args.encode_subsampling == '420':
            jpeg.CHROMA_HALVE_V_TYPE1 = 1
    elif pixfmt == 'YUV420-planar':
        if args.encode_subsampling in ['422', '444']:
            jpeg.CHROMA_DOUBLE_V = 1
    else:
        assert False

    # none of this seems to affect anything????
    jpeg.REG_0x94 = 0xc     # c/2 for 444; 8/2 for 422; 3/1 for 411; b/2 for 400
    jpeg.REG_0x98 = 0x2
    jpeg.REG_0x20c = im_W
    jpeg.REG_0x210 = im_H

    if pixfmt in ['RGB888', 'RGB101010', 'RGB565']:
        jpeg.CONVERT_COLOR_SPACE = 1
    jpeg.MATRIX_MULT[0].val = 0x4d
    jpeg.MATRIX_MULT[1].val = 0x96
    jpeg.MATRIX_MULT[2].val = 0x1d
    jpeg.MATRIX_MULT[3].val = 0xffffffd5
    jpeg.MATRIX_MULT[4].val = 0xffffffab
    jpeg.MATRIX_MULT[5].val = 0x80
    jpeg.MATRIX_MULT[6].val = 0x80
    jpeg.MATRIX_MULT[7].val = 0xffffff95
    jpeg.MATRIX_MULT[8].val = 0xffffffeb
    jpeg.MATRIX_MULT[9].val = 0x0
    jpeg.MATRIX_MULT[10].val = 0x80

    if pixfmt == 'RGB888':
        jpeg.ENCODE_PIXEL_FORMAT.set(FORMAT=E_ENCODE_PIXEL_FORMAT.RGB888)
    elif pixfmt == 'RGB101010':
        jpeg.ENCODE_PIXEL_FORMAT.set(FORMAT=E_ENCODE_PIXEL_FORMAT.RGB101010)
    elif pixfmt == 'RGB565':
        jpeg.ENCODE_PIXEL_FORMAT.set(FORMAT=E_ENCODE_PIXEL_FORMAT.RGB565)
    elif pixfmt == 'YUV10':
        jpeg.ENCODE_PIXEL_FORMAT.set(FORMAT=E_ENCODE_PIXEL_FORMAT.YUV10_linear)
    elif pixfmt == 'YUV-linear':
        jpeg.ENCODE_PIXEL_FORMAT.set(FORMAT=E_ENCODE_PIXEL_FORMAT.YUV_linear)
    elif pixfmt in ['YUV444-planar', 'YUV422-planar', 'YUV420-planar']:
        jpeg.ENCODE_PIXEL_FORMAT.set(FORMAT=E_ENCODE_PIXEL_FORMAT.YUV_planar)
    else:
        assert False
    if pixfmt == 'YUV-linear':
        jpeg.ENCODE_COMPONENT0_POS = 0
        jpeg.ENCODE_COMPONENT1_POS = 1
        jpeg.ENCODE_COMPONENT2_POS = 3
        jpeg.ENCODE_COMPONENT3_POS = 2
    elif pixfmt in ['YUV422-planar', 'YUV420-planar', 'YUV444-planar']:
        jpeg.ENCODE_COMPONENT0_POS = 0
        jpeg.ENCODE_COMPONENT1_POS = 0
        jpeg.ENCODE_COMPONENT2_POS = 1
        jpeg.ENCODE_COMPONENT3_POS = 3
    else:
        jpeg.ENCODE_COMPONENT0_POS = 0
        jpeg.ENCODE_COMPONENT1_POS = 1
        jpeg.ENCODE_COMPONENT2_POS = 2
        jpeg.ENCODE_COMPONENT3_POS = 3

    jpeg.INPUT_START1 = input_buf_iova
    jpeg.INPUT_START2 = input_buf_iova + surface_P1_off
    jpeg.INPUT_END = input_buf_iova + input_mem_sz + 7  # NOTE +7
    jpeg.OUTPUT_START1 = output_buf_iova
    jpeg.OUTPUT_START2 = 0xdeadbeef
    jpeg.OUTPUT_END = output_buf_iova + output_mem_sz

    jpeg.REG_0x118 = 0x1
    jpeg.REG_0x11c = 0x0

    jpeg.ENABLE_RST_LOGGING = args.encode_rst_interval is not None

    jpeg.MODE = 0x16f
    if args.encode_subsampling == '444':
        jpeg_subsampling = E_JPEG_IO_FLAGS_SUBSAMPLING._444
    elif args.encode_subsampling == '422':
        jpeg_subsampling = E_JPEG_IO_FLAGS_SUBSAMPLING._422
    elif args.encode_subsampling == '420':
        jpeg_subsampling = E_JPEG_IO_FLAGS_SUBSAMPLING._420
    elif args.encode_subsampling == '400':
        jpeg_subsampling = E_JPEG_IO_FLAGS_SUBSAMPLING._400
    else:
        assert False
    jpeg.JPEG_IO_FLAGS.set(
        OUTPUT_8BYTE_CHUNKS_CORRECTLY=1,
        OUTPUT_MACROBLOCKS_UNFLIPPED_H=1,
        SUBSAMPLING_MODE=jpeg_subsampling
    )
    jpeg.JPEG_WIDTH = im_W
    jpeg.JPEG_HEIGHT = im_H
    if args.encode_rst_interval is not None:
        jpeg.RST_INTERVAL = args.encode_rst_interval
    else:
        jpeg.RST_INTERVAL = 0
    jpeg.JPEG_OUTPUT_FLAGS = 0

    jpeg.QTBL[0].val =  0xa06e64a0
    jpeg.QTBL[1].val =  0xf0ffffff
    jpeg.QTBL[2].val =  0x78788cbe
    jpeg.QTBL[3].val =  0xffffffff
    jpeg.QTBL[4].val =  0x8c82a0f0
    jpeg.QTBL[5].val =  0xffffffff
    jpeg.QTBL[6].val =  0x8caadcff
    jpeg.QTBL[7].val =  0xffffffff
    jpeg.QTBL[8].val =  0xb4dcffff
    jpeg.QTBL[9].val =  0xffffffff
    jpeg.QTBL[10].val = 0xf0ffffff
    jpeg.QTBL[11].val = 0xffffffff
    jpeg.QTBL[12].val = 0xffffffff
    jpeg.QTBL[13].val = 0xffffffff
    jpeg.QTBL[14].val = 0xffffffff
    jpeg.QTBL[15].val = 0xffffffff

    jpeg.QTBL[16].val = 0xaab4f0ff
    jpeg.QTBL[17].val = 0xffffffff
    jpeg.QTBL[18].val = 0xb4d2ffff
    jpeg.QTBL[19].val = 0xffffffff
    jpeg.QTBL[20].val = 0xf0ffffff
    jpeg.QTBL[21].val = 0xffffffff
    jpeg.QTBL[22].val = 0xffffffff
    jpeg.QTBL[23].val = 0xffffffff
    jpeg.QTBL[24].val = 0xffffffff
    jpeg.QTBL[25].val = 0xffffffff
    jpeg.QTBL[26].val = 0xffffffff
    jpeg.QTBL[27].val = 0xffffffff
    jpeg.QTBL[28].val = 0xffffffff
    jpeg.QTBL[29].val = 0xffffffff
    jpeg.QTBL[30].val = 0xffffffff
    jpeg.QTBL[31].val = 0xffffffff

    jpeg.QTBL[32].val = 0x01010201
    jpeg.QTBL[33].val = 0x01020202
    jpeg.QTBL[34].val = 0x02030202
    jpeg.QTBL[35].val = 0x03030604
    jpeg.QTBL[36].val = 0x03030303
    jpeg.QTBL[37].val = 0x07050804
    jpeg.QTBL[38].val = 0x0608080a
    jpeg.QTBL[39].val = 0x0908070b
    jpeg.QTBL[40].val = 0x080a0e0d
    jpeg.QTBL[41].val = 0x0b0a0a0c
    jpeg.QTBL[42].val = 0x0a08080b
    jpeg.QTBL[43].val = 0x100c0c0d
    jpeg.QTBL[44].val = 0x0f0f0f0f
    jpeg.QTBL[45].val = 0x090b1011
    jpeg.QTBL[46].val = 0x0f0e110d
    jpeg.QTBL[47].val = 0x0e0e0e01

    jpeg.QTBL[48].val = 0x04040405
    jpeg.QTBL[49].val = 0x04050905
    jpeg.QTBL[50].val = 0x05090f0a
    jpeg.QTBL[51].val = 0x080a0f1a
    jpeg.QTBL[52].val = 0x13090913
    jpeg.QTBL[53].val = 0x1a1a1a1a
    jpeg.QTBL[54].val = 0x0d1a1a1a
    jpeg.QTBL[55].val = 0x1a1a1a1a
    jpeg.QTBL[56].val = 0x1a1a1a1a
    jpeg.QTBL[57].val = 0x1a1a1a1a
    jpeg.QTBL[58].val = 0x1a1a1a1a
    jpeg.QTBL[59].val = 0x1a1a1a1a
    jpeg.QTBL[60].val = 0x1a1a1a1a
    jpeg.QTBL[61].val = 0x1a1a1a1a
    jpeg.QTBL[62].val = 0x1a1a1a1a
    jpeg.QTBL[63].val = 0x1a1a1a1a

    jpeg.HUFFMAN_TABLE.val = 0x3c
    jpeg.QTBL_SEL.val = 0xff
    jpeg.REG_0x0.val = 0x1
    jpeg.REG_0x1004.val = 0x1

    # FIXME: we don't actually know when it's done
    time.sleep(1)

    print(jpeg.STATUS.reg)
    print(jpeg.PERFCOUNTER.reg)
    jpeg_out_sz = jpeg.COMPRESSED_BYTES.val
    print(f"JPEG output is {jpeg_out_sz} bytes")

    rst_log_n = jpeg.RST_LOG_ENTRIES.val
    for i in range(rst_log_n):
        print(f"RST log[{i}] = 0x{jpeg.RSTLOG[i].val:X}")

    output_data = iface.readmem(output_buf_phys, output_mem_sz)
    if args.raw_output is not None:
        with open(args.raw_output, 'wb') as f:
            f.write(output_data)
    with open(args.output, 'wb') as f:
        f.write(output_data[:jpeg_out_sz])



# SPDX-License-Identifier: MIT
import os, tempfile, shutil, subprocess, re
from . import sysreg

__all__ = ["AsmException", "ARMAsm"]

uname = os.uname()

if uname.sysname == "Darwin":
    DEFAULT_ARCH = "aarch64-linux-gnu-"
    if uname.machine == "arm64":
        TOOLCHAIN = "/opt/homebrew/opt/llvm/bin/"
    else:
        TOOLCHAIN = "/usr/local/opt/llvm/bin/"
    USE_CLANG = "1"
elif uname.sysname == "OpenBSD":
    DEFAULT_ARCH = "aarch64-none-elf-"
    TOOLCHAIN = "/usr/local/bin/"
    USE_CLANG = "1"
else:
    if uname.machine == "aarch64":
        DEFAULT_ARCH = ""
    else:
        DEFAULT_ARCH = "aarch64-linux-gnu-"
    USE_CLANG = "0"
    TOOLCHAIN = ""

use_clang = os.environ.get("USE_CLANG", USE_CLANG).strip() == "1"
toolchain = os.environ.get("TOOLCHAIN", TOOLCHAIN)

if use_clang:
    CC = toolchain + "clang --target=%ARCH"
    LD = toolchain + "ld.lld"
    OBJCOPY = toolchain + "llvm-objcopy"
    OBJDUMP = toolchain + "llvm-objdump"
    NM = toolchain + "llvm-nm"
else:
    CC = toolchain + "%ARCHgcc"
    LD = toolchain + "%ARCHld"
    OBJCOPY = toolchain + "%ARCHobjcopy"
    OBJDUMP = toolchain + "%ARCHobjdump"
    NM = toolchain + "%ARCHnm"

class AsmException(Exception):
    pass

class BaseAsm(object):
    def __init__(self, source, addr = 0):
        self.source = source
        self._tmp = tempfile.mkdtemp() + os.sep
        self.addr = addr
        self.compile(source)

    def _call(self, program, args):
        subprocess.check_call(program.replace("%ARCH", self.ARCH) + " " + args, shell=True)

    def _get(self, program, args):
        return subprocess.check_output(program.replace("%ARCH", self.ARCH) + " " + args, shell=True).decode("ascii")

    def compile(self, source):
        for name, enc in sysreg.sysreg_fwd.items():
            source = re.sub("\\b" + name + "\\b", f"s{enc[0]}_{enc[1]}_c{enc[2]}_c{enc[3]}_{enc[4]}", source)

        self.sfile = self._tmp + "b.S"
        with open(self.sfile, "w") as fd:
            fd.write(self.HEADER + "\n")
            fd.write(source + "\n")
            fd.write(self.FOOTER + "\n")

        self.ofile = self._tmp + "b.o"
        self.elffile = self._tmp + "b.elf"
        self.bfile = self._tmp + "b.b"
        self.nfile = self._tmp + "b.n"

        self._call(CC, f"{self.CFLAGS} -c -o {self.ofile} {self.sfile}")
        self._call(LD, f"{self.LDFLAGS} --Ttext={self.addr:#x} -o {self.elffile} {self.ofile}")
        self._call(OBJCOPY, f"-j.text -O binary {self.elffile} {self.bfile}")
        self._call(NM, f"{self.elffile} > {self.nfile}")

        with open(self.bfile, "rb") as fd:
            self.data = fd.read()

        with open(self.nfile) as fd:
            for line in fd:
                line = line.replace("\n", "")
                addr, type, name = line.split()
                addr = int(addr, 16)
                setattr(self, name, addr)
        self.start = self._start
        self.len = len(self.data)
        self.end = self.start + self.len

    def objdump(self):
        self._call(OBJDUMP, f"-rd {self.elffile}")

    def disassemble(self):
        output = self._get(OBJDUMP, f"-zd {self.elffile}")

        for line in output.split("\n"):
            if not line or line.startswith("/"):
                continue
            sl = line.split()
            if not sl or sl[0][-1] != ":":
                continue
            yield line

    def __del__(self):
        if self._tmp:
            shutil.rmtree(self._tmp)
            self._tmp = None

class ARMAsm(BaseAsm):
    ARCH = os.path.join(os.environ.get("ARCH", DEFAULT_ARCH))
    CFLAGS = "-pipe -Wall -march=armv8.4-a"
    if use_clang:
        LDFLAGS = "-maarch64elf"
    else:
        LDFLAGS = "-maarch64linux"
    HEADER = """
    .text
    .globl _start
_start:
    """
    FOOTER = """
    .pool
    """

if __name__ == "__main__":
    import sys
    code = """
    ldr x0, =0xDEADBEEF
    b test
    mrs x0, spsel
    svc 1
    %s
test:
    b test
    ret
""" % (" ".join(sys.argv[1:]))
    c = ARMAsm(code, 0x1238)
    c.objdump()
    assert c.start == 0x1238
    if not sys.argv[1:]:
        assert c.test == 0x1248



# SPDX-License-Identifier: MIT

from m1n1.trace import Tracer
from m1n1.trace.dart import DARTTracer
from m1n1.trace.asc import ASCTracer, EP, EPState, msg, msg_log, DIR, EPContainer
from m1n1.utils import *
from m1n1.constructutils import *
from m1n1.fw.afk.rbep import *
from m1n1.fw.afk.epic import *
from m1n1.fw.aop import *
from m1n1.fw.aop.ipc import *

import sys

class AFKRingBufSniffer(AFKRingBuf):
    def __init__(self, ep, state, base, size):
        super().__init__(ep, base, size)
        self.state = state
        self.rptr = getattr(state, "rptr", 0)

    def update_rptr(self, rptr):
        self.state.rptr = rptr

    def update_wptr(self):
        raise NotImplementedError()

    def get_wptr(self):
        return struct.unpack("<I", self.read_buf(2 * self.BLOCK_SIZE, 4))[0]

    def read_buf(self, off, size):
        return self.ep.dart.ioread(0, self.base + off, size)

class AFKEp(EP):
    BASE_MESSAGE = AFKEPMessage

    def __init__(self, tracer, epid):
        super().__init__(tracer, epid)
        self.txbuf = None
        self.rxbuf = None
        self.state.txbuf = EPState()
        self.state.rxbuf = EPState()
        self.state.shmem_iova = None
        self.state.txbuf_info = None
        self.state.rxbuf_info = None
        self.state.verbose = 1

    def start(self):
        self.create_bufs()

    def create_bufs(self):
        if not self.state.shmem_iova:
            return
        if not self.txbuf and self.state.txbuf_info:
            off, size = self.state.txbuf_info
            self.txbuf = AFKRingBufSniffer(self, self.state.txbuf,
                                           self.state.shmem_iova + off, size)
        if not self.rxbuf and self.state.rxbuf_info:
            off, size = self.state.rxbuf_info
            self.rxbuf = AFKRingBufSniffer(self, self.state.rxbuf,
                                           self.state.shmem_iova + off, size)

    Init =          msg_log(0x80, DIR.TX)
    Init_Ack =      msg_log(0xa0, DIR.RX)

    GetBuf =        msg_log(0x89, DIR.RX)

    Shutdown =      msg_log(0xc0, DIR.TX)
    Shutdown_Ack =  msg_log(0xc1, DIR.RX)

    @msg(0xa1, DIR.TX, AFKEP_GetBuf_Ack)
    def GetBuf_Ack(self, msg):
        self.state.shmem_iova = msg.DVA
        self.txbuf = None
        self.rxbuf = None
        self.state.txbuf = EPState()
        self.state.rxbuf = EPState()
        self.state.txbuf_info = None
        self.state.rxbuf_info = None

    @msg(0xa2, DIR.TX, AFKEP_Send)
    def Send(self, msg):
        for data in self.txbuf.read():
            #if self.state.verbose >= 3:
            if True:
                self.log(f"===TX DATA=== epid={self.epid} rptr={self.txbuf.state.rptr:#x}")
                chexdump(data)
                self.log(f"===END DATA===")
                self.log("Backtrace on TX data:")
                self.hv.bt()
            self.handle_ipc(data, dir=">")
        return True

    Hello =         msg_log(0xa3, DIR.TX)

    @msg(0x85, DIR.RX, AFKEPMessage)
    def Recv(self, msg):
        for data in self.rxbuf.read():
            #if self.state.verbose >= 3:
            if True:
                self.log(f"===RX DATA=== epid={self.epid} rptr={self.rxbuf.state.rptr:#x}")
                chexdump(data)
                self.log(f"===END DATA===")
            self.handle_ipc(data, dir="<")
        return True

    def handle_ipc(self, data, dir=None):
        pass

    @msg(0x8a, DIR.RX, AFKEP_InitRB)
    def InitTX(self, msg):
        off = msg.OFFSET * AFKRingBuf.BLOCK_STEP
        size = msg.SIZE * AFKRingBuf.BLOCK_STEP
        self.state.txbuf_info = (off, size)
        self.create_bufs()

    @msg(0x8b, DIR.RX, AFKEP_InitRB)
    def InitRX(self, msg):
        off = msg.OFFSET * AFKRingBuf.BLOCK_STEP
        size = msg.SIZE * AFKRingBuf.BLOCK_STEP
        self.state.rxbuf_info = (off, size)
        self.create_bufs()

class DummyAFKEp(AFKEp):
    def handle_ipc(self, data, dir=None):
        pass

class EPICEp(AFKEp):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pending_call = None
        self.pending_cmd = None

    def handle_hello(self, hdr, sub, fd):
        if sub.type != 0xc0:
            return False

        payload = fd.read()
        name = payload.split(b"\0")[0].decode("ascii")
        self.log(f"Hello! (endpoint {name})")
        return True

    def handle_notify(self, hdr, sub, fd):
        for calltype in CALLTYPES:
            if calltype.matches(hdr, sub):
                call = calltype.from_stream(fd)
                self.trace_call_early(call)
                self.pending_call = call
                return True

        return False

    def handle_reply(self, hdr, sub, fd):
        if self.pending_call is None:
            return False

        call = self.pending_call
        call.read_resp(fd)
        self.trace_call(call)
        self.pending_call = None
        return True

    def dispatch_ipc(self, dir, hdr, sub, fd):
        if sub.category == EPICCategory.COMMAND:
            return self.handle_notify(hdr, sub, fd)
        if dir == "<" and sub.category == EPICCategory.REPORT:
            return self.handle_hello(hdr, sub, fd)
        if dir == ">" and sub.category == EPICCategory.NOTIFY:
            return self.handle_notify(hdr, sub, fd)
        if dir == "<" and sub.category == EPICCategory.REPLY:
            return self.handle_reply(hdr, sub, fd)

    def handle_ipc(self, data, dir=None):
        fd = BytesIO(data)
        hdr = EPICHeader.parse_stream(fd)
        sub = EPICSubHeaderVer2.parse_stream(fd)

        if not getattr(self, 'VERBOSE', False):
            return

        self.log(f"{dir} 0x{hdr.channel:x} Type {hdr.type} Ver {hdr.version} Tag {hdr.seq}")
        self.log(f"  Len {sub.length} Ver {sub.version} Cat {sub.category} Type {sub.type:#x} Ts {sub.timestamp:#x}")
        self.log(f"  Unk1 {sub.unk1:#x} Unk2 {sub.unk2:#x}")

        if self.dispatch_ipc(dir, hdr, sub, fd):
            return

    def trace_call_early(self, call):
        # called at TX time
        if isinstance(call, IndirectCall):
            call.read_txbuf(self)

    def trace_call(self, call):
        if isinstance(call, IndirectCall):
            call.read_rxbuf(self)
            call = call.unwrap()
        call.dump(self.log)

class SPUAppEp(EPICEp):
    SHORT = "SPUApp"

class AccelEp(EPICEp):
    SHORT = "accel"

class GyroEp(EPICEp):
    SHORT = "gyro"

class LASEp(EPICEp):
    SHORT = "las"

class WakeHintEp(EPICEp):
    SHORT = "wakehint"

class UNK26Ep(EPICEp):
    SHORT = "unk26"

class AudioEp(EPICEp):
    SHORT = "aop-audio"
    VERBOSE = True

class VoiceTriggerEp(EPICEp):
    SHORT = "aop-voicetrigger"
    VERBOSE = True


class AOPTracer(ASCTracer, AOPBase):
    ENDPOINTS = {
        0x20: SPUAppEp,
        0x21: AccelEp,
        0x22: GyroEp,
        0x24: LASEp,
        0x25: WakeHintEp,
        0x26: UNK26Ep,
        0x27: AudioEp,
        0x28: VoiceTriggerEp,
    }

    def __init__(self, hv, devpath, verbose=False):
        self.default_bootargs = None
        super().__init__(hv, devpath, verbose)
        self.u = hv.u
        AOPBase.__init__(self, hv.u, self.dev)

    def start(self, *args):
        self.default_bootargs = self.read_bootargs()
        super().start(*args)

    def w_CPU_CONTROL(self, val):
        if val.RUN:
            self.bootargs = self.read_bootargs()
            self.log("Bootargs patched by AP:")
            self.default_bootargs.dump_diff(self.bootargs, self.log)
            self.log("(End of list)")
        super().w_CPU_CONTROL(val)

    @classmethod
    def replay(cls, f, passthru=False):
        epmap = dict()
        epcont = EPContainer()

        class FakeASCTracer:
            def __init__(self):
                self.hv = None

            def log(self, str):
                print(str)
        asc_tracer = FakeASCTracer()

        for cls in cls.mro():
            eps = getattr(cls, "ENDPOINTS", None)
            if eps is None:
                break
            for k, v in eps.items():
                if k in epmap:
                    continue
                ep = v(asc_tracer, k)
                epmap[k] = ep
                if getattr(epcont, ep.name, None):
                    ep.name = f"{ep.name}{k:02x}"
                setattr(epcont, ep.name, ep)
                ep.start()

        def readdump(firstline, hdr, f):
            l = firstline
            assert hdr in l
            postscribe = l[l.index(hdr) + len(hdr):]
            annotation = dict([s.split("=") for s \
                              in postscribe.strip().split(" ")])

            dump = []
            for l in f:
                if "===END DATA===" in l:
                    break
                dump.append(l)
            return chexundump("".join(dump)), annotation

        def read_txbuf(icall, ep):
            hdr = "===COMMAND TX DATA==="
            for l in f:
                if hdr in l:
                    break
            data, annot = readdump(l, hdr, f)
            assert int(annot["addr"], 16) == icall.args.txbuf
            icall.txbuf = data
        def read_rxbuf(icall, ep):
            hdr = "===COMMAND RX DATA==="
            for l in f:
                if hdr in l:
                    break
            data, annot = readdump(l, hdr, f)
            assert int(annot["addr"], 16) == icall.rets.rxbuf
            icall.rxbuf = data
        IndirectCall.read_rxbuf = read_rxbuf
        IndirectCall.read_txbuf = read_txbuf

        for l in f:
            if (rxhdr := "===RX DATA===") in l:
                dir = "<"
                hdr = rxhdr
            elif (txhdr := "===TX DATA===") in l:
                dir = ">"
                hdr = txhdr
            else:
                if passthru:
                    print(l, end="")
                continue
            data, annot = readdump(l, hdr, f)
            epid = int(annot["epid"])
            epmap[epid].handle_ipc(data, dir)
                        

if __name__ == "__main__":
    # We can replay traces by saving the textual output of live tracing
    # and then passing it to this script.
    with open(sys.argv[1]) as f:
        AOPTracer.replay(f)
    sys.exit(0)

dart_aop_tracer = DARTTracer(hv, "/arm-io/dart-aop", verbose=4)
dart_aop_tracer.start()

dart_aop_base = u.adt["/arm-io/dart-aop"].get_reg(0)[0]

#hv.trace_range(irange(*u.adt["/arm-io/dart-aop"].get_reg(1)))
#hv.trace_range(irange(*u.adt["/arm-io/aop"].get_reg(1)))
#hv.trace_range(irange(*u.adt["/arm-io/aop"].get_reg(3)))
#hv.trace_range(irange(*u.adt["/arm-io/admac-aop-audio"].get_reg(0)))

aop_tracer = AOPTracer(hv, "/arm-io/aop", verbose=1)
aop_tracer.start(dart_aop_tracer.dart)



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, pathlib, traceback
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

import argparse, pathlib
from io import BytesIO

def volumespec(s):
    return tuple(s.split(":", 2))

parser = argparse.ArgumentParser(description='Run a Mach-O payload under the hypervisor')
parser.add_argument('-s', '--symbols', type=pathlib.Path)
parser.add_argument('-m', '--script', type=pathlib.Path, action='append', default=[])
parser.add_argument('-c', '--command', action="append", default=[])
parser.add_argument('-S', '--shell', action="store_true")
parser.add_argument('-e', '--hook-exceptions', action="store_true")
parser.add_argument('-d', '--debug-xnu', action="store_true")
parser.add_argument('-l', '--logfile', type=pathlib.Path)
parser.add_argument('-C', '--cpus', default=None)
parser.add_argument('-r', '--raw', action="store_true")
parser.add_argument('-E', '--entry-point', action="store", type=int, help="Entry point for the raw image", default=0x800)
parser.add_argument('-a', '--append-payload', type=pathlib.Path, action="append", default=[])
parser.add_argument('-v', '--volume', type=volumespec, action='append',
                    help='Attach a 9P virtio device for file export to the guest. The argument is a host path to the '
                         'exported tree, joined by colon (\':\') with a tag under which the tree will be advertised '
                         'on the guest side.')
parser.add_argument('payload', type=pathlib.Path)
parser.add_argument('boot_args', default=[], nargs="*")
args = parser.parse_args()

from m1n1.proxy import *
from m1n1.proxyutils import *
from m1n1.utils import *
from m1n1.shell import run_shell
from m1n1.hv import HV
from m1n1.hv.virtio import Virtio9PTransport
from m1n1.hw.pmu import PMU

iface = UartInterface()
p = M1N1Proxy(iface, debug=False)
bootstrap_port(iface, p)
u = ProxyUtils(p, heap_size = 128 * 1024 * 1024)

hv = HV(iface, p, u)

hv.hook_exceptions = args.hook_exceptions

hv.init()

if args.cpus:
    avail = [i.name for i in hv.adt["/cpus"]]
    want = set(f"cpu{i}" for i in args.cpus)
    for cpu in avail:
        if cpu in want:
            continue
        try:
            del hv.adt[f"/cpus/{cpu}"]
            print(f"Disabled {cpu}")
        except KeyError:
            continue

if args.debug_xnu:
    hv.adt["chosen"].debug_enabled = 1

if args.volume:
    for path, tag in args.volume:
        hv.attach_virtio(Virtio9PTransport(root=path, tag=tag))

if args.logfile:
    hv.set_logfile(args.logfile.open("w"))

if len(args.boot_args) > 0:
    boot_args = " ".join(args.boot_args)
    hv.set_bootargs(boot_args)

symfile = None
if args.symbols:
    symfile = args.symbols.open("rb")

payload = args.payload.open("rb")

if args.append_payload:
    concat = BytesIO()
    concat.write(payload.read())
    for part in args.append_payload:
        concat.write(part.open("rb").read())
    concat.seek(0)
    payload = concat

if args.raw:
    hv.load_raw(payload.read(), args.entry_point)
else:
    hv.load_macho(payload, symfile=symfile)

PMU(u).reset_panic_counter()

for i in args.script:
    try:
        hv.run_script(i)
    except:
        traceback.print_exc()
        args.shell = True

for i in args.command:
    try:
        hv.run_code(i)
    except:
        traceback.print_exc()
        args.shell = True

if args.shell:
    run_shell(hv.shell_locals, "Entering hypervisor shell. Type ^D to start the guest.")

hv.start()

run_shell(hv.shell_locals, "Hypervisor exited. Entering shell.")

p.smp_stop_secondaries(True)
p.sleep(True)



import json, sys
import argparse

parser = argparse.ArgumentParser()
parser.add_argument("--imp-apl-prefix", action="store_true")
parser.add_argument("regfile")
args = parser.parse_args()

if args.imp_apl_prefix:
    prefix = "IMP_APL_"
else:
    prefix = ""

data = json.load(open(args.regfile))
for reg in data:
    name = reg['name']

    print(f"#define SYS_{prefix}{name} sys_reg({', '.join(str(i) for i in reg['enc'])})")

    if name[-4:-1] == "_EL":
        name = name[:-4]

    for fieldset in reg.get("fieldsets", []):
        if "instance" in fieldset:
            print(f"// {fieldset['instance']}")
        for f in fieldset["fields"]:
            fname = f["name"]
            msb, lsb = f["msb"], f["lsb"]

            if msb == lsb:
                print(f"#define {name}_{fname} BIT({lsb})")
            else:
                print(f"#define {name}_{fname} GENMASK({msb}, {lsb})")

    print()



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, pathlib
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

from contextlib import contextmanager

from m1n1.setup import *
from m1n1.find_regs import *
from m1n1 import asm

p.smp_start_secondaries()

class ARMPageTable:
    PAGESIZE = 0x4000

    def __init__(self, memalign, free):
        self.memalign = memalign
        self.free = free

        self.l0 = self.memalign(self.PAGESIZE, self.PAGESIZE)
        self.l1 = [self.memalign(self.PAGESIZE, self.PAGESIZE), self.memalign(
            self.PAGESIZE, self.PAGESIZE)]
        self.l2 = {}

        p.write64(self.l0, self.make_table_pte(self.l1[0]))
        p.write64(self.l0+8, self.make_table_pte(self.l1[1]))

    def make_table_pte(self, addr):
        # table mapping, access bit set
        return addr | 0b11 | (1 << 10)

    def map_page(self, vaddr, paddr, access_bits):
        ap = (access_bits & 0b1100) >> 2
        pxn = (access_bits & 0b0010) >> 1
        uxn = (access_bits & 0b0001)

        # block mapping, access bit set
        pte = paddr | 0b01 | (1 << 10)

        # move access bits in place
        pte |= ap << 6
        pte |= pxn << 54
        pte |= uxn << 53

        l0_idx = (vaddr >> (25+11+11)) & 1
        l1_idx = (vaddr >> (25+11)) & 0x7ff
        l2_idx = (vaddr >> 25) & 0x7ff

        tbl = self.l2.get((l0_idx, l1_idx), None)
        if not tbl:
            tbl = self.memalign(self.PAGESIZE, self.PAGESIZE)
            self.l2[(l0_idx, l1_idx)] = tbl
            p.write64(self.l1[l0_idx] + 8*l1_idx, self.make_table_pte(tbl))

        p.write64(tbl + 8*l2_idx, pte)

    def map(self, vaddr, paddr, sz, access_bits):
        assert sz % 0x2000000 == 0
        assert vaddr % 0x2000000 == 0
        assert paddr % 0x2000000 == 0
        assert access_bits <= 0b1111

        while sz > 0:
            self.map_page(vaddr, paddr, access_bits)
            sz -= 0x2000000
            vaddr += 0x2000000
            paddr += 0x2000000


def build_and_write_code(heap, code):
    page = heap.memalign(0x4000, 0x4000)
    compiled = asm.ARMAsm(code, page).data
    iface.writemem(page, compiled)
    p.dc_cvau(page, len(compiled))
    p.ic_ivau(page, len(compiled))
    return page


def setup_exception_vectors(heap, gxf=False):
    if gxf:
        elr = "S3_6_C15_C10_6"
        eret = ".long 0x201400"
        indicator = 0xf2
    else:
        elr = "ELR_EL1"
        eret = "eret"
        indicator = 0xf0

    return build_and_write_code(heap, """
                .rept   16
                b 1f
                .align 7
                .endr

                1:
                // store that we failed
                mov x10, 0x{indicator:x}

                // move PC two instruction further and clear 0xf0 0000 0000 to
                // make sure we end up in the r-x mapping either way and don't
                // repeat the instruction that just faulted
                // we skip the second instruction since that one is used to
                // indicate success
                ldr x11, =0x0fffffffff
                mrs x12, {elr}
                add x12, x12, #8
                and x12, x12, x11
                msr {elr}, x12

                isb
                {eret}
        """.format(eret=eret, elr=elr, indicator=indicator))


print("Setting up..")
pagetable = ARMPageTable(u.memalign, u.free)
pagetable.map(0x800000000, 0x800000000, 0xc00000000, 0)
pagetable.map(0xf800000000, 0x800000000, 0xc00000000, 1)

el2_vectors = setup_exception_vectors(u.heap, gxf=False)
gl2_vectors = setup_exception_vectors(u.heap, gxf=True)

probe_page = build_and_write_code(u.heap, "mov x10, 0x80\nret\nret\nret\n")
probe_page_vaddr = probe_page | 0xf000000000

code_page = build_and_write_code(u.heap, """
            #define SPRR_PERM_EL0  S3_6_C15_C1_5
            #define SPRR_PERM_EL1  S3_6_C15_C1_6
            #define SPRR_PERM_EL2  S3_6_C15_C1_7

            #define GXF_CONFIG_EL2 s3_6_c15_c1_2
            #define GXF_ENTER_EL2 s3_6_c15_c8_1
            #define GXF_ABORT_EL2 s3_6_c15_c8_2

            #define MPIDR_GL2 S3_6_C15_C10_1
            #define VBAR_GL2 S3_6_C15_C10_2
            #define SPSR_GL2 S3_6_C15_C10_3
            #define ELR_GL2 S3_6_C15_C10_6
            #define FAR_GL2 S3_6_C15_C10_7

            #define genter .long 0x00201420
            #define gexit .long 0x201400

            // just store everything since i'm too lazy to think about
            // register assignments
            str x30, [sp, #-16]!
            stp x28, x29, [sp, #-16]!
            stp x26, x27, [sp, #-16]!
            stp x24, x25, [sp, #-16]!
            stp x22, x23, [sp, #-16]!
            stp x20, x21, [sp, #-16]!
            stp x18, x19, [sp, #-16]!
            stp x16, x17, [sp, #-16]!
            stp x14, x15, [sp, #-16]!
            stp x12, x13, [sp, #-16]!
            stp x10, x11, [sp, #-16]!
            stp x8, x9, [sp, #-16]!
            stp x6, x7, [sp, #-16]!
            stp x4, x5, [sp, #-16]!
            stp x2, x3, [sp, #-16]!
            stp x0, x1, [sp, #-16]!

            mov x20, x0  // store SPRR value for later
            mov x21, 0   // clear result

            // setup exception vectors
            ldr x0, =0x{vectors:x}
            msr VBAR_EL2, x0
            isb

            // prepare MMU
            ldr x0, =0x0400ff
            msr MAIR_EL1, x0
            ldr x0, =0x27510b510
            msr TCR_EL1, x0
            ldr x0, =0x{ttbr:x}
            msr TTBR0_EL2, x0

            // enable SPPR
            mov x0, 1
            msr s3_6_c15_c1_0, x0
            msr s3_6_c15_c1_3, xzr
            isb

            // clear all SPPR registers
            // (note that reads from/writes to EL1 will be redirected to EL2 anyway)
            ldr x0, =0
            msr SPRR_PERM_EL0, x0
            msr SPRR_PERM_EL1, x0
            msr SPRR_PERM_EL2, x0
            msr s3_6_c15_c1_3, x0

            // setup SPPR_EL2
            msr SPRR_PERM_EL2, x20
            isb
            dsb ishst
            tlbi vmalle1is
            dsb ish
            isb


            msr s3_6_c15_c1_3, xzr
            isb

            // enable MMU
            ldr x1, =0x1005
            mrs x0, SCTLR_EL1
            mov x3, x0
            orr x0, x0, x1
            msr SCTLR_EL1, x0
            isb

            // configure and enable GXF
            mov x0, 1
            msr GXF_CONFIG_EL2, x0
            isb
            ldr x0, =gxf_entry
            msr GXF_ENTER_EL2, x0
            ldr x0, =gxf_abort
            msr GXF_ABORT_EL2, x0
            isb

            // test GXF access
            genter

            // test execute access
            ldr x1, =0x{probe_page:x}
            mov x10, 0
            blr x1
            lsl x21, x21, #8
            orr x21, x21, x10

            // test read access
            ldr x1, =0x{probe_page:x}
            mov x10, 0
            ldr x1, [x1]
            mov x10, 0x80
            lsl x21, x21, #8
            orr x21, x21, x10

            // test write access
            ldr x1, =0x{probe_page:x}
            add x1, x1, 0x20
            mov x10, 0
            str x1, [x1]
            mov x10, 0x80
            lsl x21, x21, #8
            orr x21, x21, x10

            // disable MMU again
            dsb ish
            isb
            msr SCTLR_EL1, x3
            isb

            mov x0, 0
            msr GXF_CONFIG_EL2, x0
            msr s3_6_c15_c1_0, x0

            mov x0, x21

            // restore everything except for x0
            add sp, sp, #8
            ldr x1, [sp], #8
            ldp x2, x3, [sp], #16
            ldp x4, x5, [sp], #16
            ldp x6, x7, [sp], #16
            ldp x8, x9, [sp], #16
            ldp x10, x11, [sp], #16
            ldp x12, x13, [sp], #16
            ldp x14, x15, [sp], #16
            ldp x16, x17, [sp], #16
            ldp x18, x19, [sp], #16
            ldp x20, x21, [sp], #16
            ldp x22, x23, [sp], #16
            ldp x24, x25, [sp], #16
            ldp x26, x27, [sp], #16
            ldp x28, x29, [sp], #16
            ldr x30, [sp], #16

            ret


            gxf_entry:
                    // setup GL exception vectors
                    ldr x0, =0x{gxf_vectors:x}
                    msr VBAR_GL2, x0
                    isb

                    // we might double fault -> store state here
                    mrs x14, S3_6_C15_C10_3
                    mrs x15, S3_6_C15_C10_4
                    mrs x16, S3_6_C15_C10_5
                    mrs x17, ELR_GL2
                    mrs x18, FAR_GL2

                    // test execute access
                    ldr x1, =0x{probe_page:x}
                    mov x10, 0
                    blr x1
                    lsl x21, x21, #8
                    orr x21, x21, x10

                    // test read access
                    ldr x1, =0x{probe_page:x}
                    mov x10, 0
                    ldr x1, [x1]
                    mov x10, 0x80
                    lsl x21, x21, #8
                    orr x21, x21, x10

                    // test write access
                    ldr x1, =0x{probe_page:x}
                    add x1, x1, #0x20
                    mov x10, 0
                    str x1, [x1]
                    mov x10, 0x80
                    lsl x21, x21, #8
                    orr x21, x21, x10

                    // restore state in case we faulted in here
                    msr S3_6_C15_C10_3, x14
                    msr S3_6_C15_C10_4, x15
                    msr S3_6_C15_C10_5, x16
                    msr ELR_GL2, x17
                    msr FAR_GL2, x18

                    isb
                    gexit

            gxf_abort:
                    // store that we failed
                    mov x10, 0xf1

                    // move PC two instruction further and clear 0xf0 0000 0000 to
                    // make sure we end up in the r-x mapping either way and don't
                    // repeat the instruction that just faulted
                    // we skip the second instruction since that one is used to
                    // indicate success
                    ldr x11, =0x0fffffffff
                    mrs x12, ELR_GL2
                    add x12, x12, #8
                    and x12, x12, x11
                    msr ELR_GL2, x12

                    isb
                    gexit
    """.format(ttbr=pagetable.l0, vectors=el2_vectors, probe_page=probe_page_vaddr, gxf_vectors=gl2_vectors))

print("Running code now...")
for i in range(0x10):
    sprr_val = 0x5 | ((i & 0xf) << 4)
    ret = p.smp_call_sync(1, code_page, sprr_val)

    glret = ret >> 24
    glx = 'x' if (glret >> 16) & 0xff == 0x80 else '-'
    glr = 'r' if (glret >> 8) & 0xff == 0x80 else '-'
    glw = 'w' if glret & 0xff == 0x80 else '-'

    x = 'x' if (ret >> 16) & 0xff == 0x80 else '-'
    r = 'r' if (ret >> 8) & 0xff == 0x80 else '-'
    w = 'w' if ret & 0xff == 0x80 else '-'

    print("SPRR: {0:04b} result: {1:x} GL: {2}{3}{4} EL: {5}{6}{7}".format(
        i, ret, glr, glw, glx, r, w, x))
